#!/usr/bin/env python
"""
Script to analyze/export/visualize the crystal structure saved in the netcdf files produced by ABINIT.
"""
from __future__ import annotations

import sys
import os
import argparse
import numpy as np
import abipy.tools.cli_parsers as cli

from pprint import pprint
from tabulate import tabulate
from monty.string import marquee
from monty.functools import prof_main
from monty.termcolor import cprint
from pymatgen.io.vasp.outputs import Xdatcar
from abipy import abilab
from abipy.core.symmetries import AbinitSpaceGroup
from abipy.core.kpoints import Ktables, Kpoint, IrredZone
from abipy.core.structure import diff_structures
from abipy.iotools.visualizer import Visualizer
from abipy.iotools.xsf import xsf_write_structure
from abipy.abio import factories


def save_structure(structure, options) -> None:
    """Save structure to file."""
    if not options.savefile: return
    print("Saving structure to file:", options.savefile)
    if os.path.exists(options.savefile):
        backup = options.savefile + ".bkp"
        print("%s already exists. Saving backup copy to: %s" % (options.savefile, backup))
        os.rename(options.savefile, backup)

    structure.to(filename=options.savefile)


def check_ordered_structure(structure) -> None:
    """Print a warning and sys.exit 1 if structure is disordered."""
    if not structure.is_ordered:
        cprint("""
Cannot handle disordered structure with fractional site occupancies.
Use OrderDisorderedStructureTransformation or EnumerateStructureTransformation
to build an appropriate supercell from partial occupancies.""", color="magenta")
        sys.exit(1)


def get_epilog() -> str:
    return """\
Usage example:

###################
# Space group tools
###################

  abistruct.py spglib FILE                 => Read structure from FILE and analyze it with spglib.
  abistruct.py abispg FILE                 => Read structure from FILE, and compute ABINIT space group.
  abistruct.py primitive FILE              => Read structure from FILE, use pymatgen and spglib to find primitive structure.
  abistruct.py abisanitize FILE            => Read structure from FILE, call abisanitize, compare structures
                                              and save "abisanitized" structure to file.
  abistruct.py conventional FILE           => Read structure from FILE, generate conventional structure
                                              following doi:10.1016/j.commatsci.2010.05.010
  abistruct.py proto FILE                  => Read structure from FILE, find possible crystallographic prototypes
                                              in the AFLOW library of crystallographic prototypes.
                                              See http://doi.org/10.1016/j.commatsci.2017.01.017

##################
# Conversion tools
##################

  abistruct.py convert FILE                => Print the ABINIT variables defining the structure.
  abistruct.py convert FILE -f cif         => Read structure from FILE and output CIF file
                                              (Use convert --help to get list of formats supported)
  abistruct.py convert out_HIST.nc         => Read FINAL structure from the HIST file and
                                              print the corresponding ABINIT variables.
  abistruct.py supercell FILE -s 2 2 1     => Read structure from FILE and build [2, 2, 1] supercell,
                                              print new structure using --format (default abivars).
################
# K-points tools
################

  abistruct.py kpath FILE                  => Read structure from FILE and print ABINIT variables for k-path.
  abistruct.py bz FILE                     => Read structure from FILE, plot BZ with matplotlib.
  abistruct.py ngkpt FILE -n 4             => Compute `ngkpt` and `shiftk` from the number of divisions used to sample
                                              the smallest reciprocal lattice vector.
  abistruct.py abikmesh FILE --ngkpt 2 2 2 => Read structure from FILE, call Abinit to sample the BZ
                                              with a 2, 2, 2 k-mesh, print points in IBZ with weights.
  abistruct.py ktables FILE -m 2 2 2       => Read structure from FILE, call spglib to sample the BZ
                                              with a 2, 2, 2 k-mesh, print points in IBZ with weights.
  abistruct.py lgk FILE -k 0.25 0 0        => Read structure from FILE, find little group of k-point,
                                              print Bilbao character table.
  abistruct.py kstar FILE -k 0.25 0 0      => Read structure from FILE, print star of k-point.
  abistruct.py keq FILE -k 0.5 0 0 0 0.5 0  => Read structure from FILE, test whether k1 and k2 are
                                               symmetry-equivalent k-points.

###############
# Miscelleanous
###############

  abistruct.py neighbors FILE              => Get neighbors for each atom in the unit cell, out to a distance radius.
  abistruct.py interpolate FILE1 FILE2     => Interpolate between two structures. Useful for the construction of NEB inputs.
  abistruct.py xrd FILE                    => X-ray diffraction plot.
  abistruct.py oxistate FILE               => Estimate oxidation states with pymatgen bond valence analysis.

###############
# Visualization
###############

  abistruct.py visualize FILE -a vesta     => Visualize the structure with vesta (default if -a is not given)
                                              Supports also ovito, xcrysden, vtk, mayavi, matplotlib See --help
  abistruct.py ipython FILE                => Read structure from FILE and open it in the Ipython terminal.
  abistruct.py notebook FILE               => Read structure from FILE and generate jupyter notebook.
  abistruct.py panel FILE                  => Generate GUI in web browser to interact with the structure
                                              Requires panel package.

###########
# Databases
###########

  abistruct.py cod_id 1526507              => Get structure from COD database (http://www.crystallography.net/cod).
  abistruct.py cod_search MgB2             => Search for structures in the COD database.
  abistruct.py mp_id mp-149                => Get structure from materials project database and print
                                              JSON representation. Use e.g. `-f abivars` to change format.
  abistruct.py mp_match FILE               => Read structure from FILE and find matching structures on the
                                              Materials Project site. Use e.g. `-f cif` to change output format.
  abistruct.py mp_search LiF               => Connect to the materials project database. Get structures corresponding
                                              to a chemical system or formula e.g. `Fe2O3` or `Li-Fe-O` or
                                              `Ir-O-*` for wildcard pattern matching.
                                              Print info and Abinit input files. Use e.g. `-f POSCAR`
                                              to change output format. `-f None` to disable structure output.
  abistruct.py mp_pd FILE-or-elements      => Generate phase diagram with entries from the Materials Project.
  abistruct.py mp_ebands FILE             => Fetch electron band structure from MP database. Print gaps.
                                             Accept FILE with structure if ebands from structure is wanted
                                             or mp id e.g. "mp-149 or list of elements e.g `Li-Fe-O` or chemical formula.

`FILE` is any file supported by abipy/pymatgen e.g Netcdf files, Abinit input/output, POSCAR, xsf ...
Use `abistruct.py --help` for help and `abistruct.py COMMAND --help` to get the documentation for `COMMAND`.
Use `-v` to increase verbosity level (can be supplied multiple times e.g -vv).
"""


def get_parser(with_epilog=False):

    # Parent parser for commands that need to know the filepath
    path_selector = argparse.ArgumentParser(add_help=False)
    path_selector.add_argument('filepath', nargs="?",
        help="File with the crystalline structure (Abinit Netcdf files, CIF, Abinit input/output files, POSCAR ...)")

    # Parent parser for commands supporting (jupyter notebooks)
    nb_parser = argparse.ArgumentParser(add_help=False)
    nb_parser.add_argument('-nb', '--notebook', default=False, action="store_true", help='Generate jupyter notebook.')
    nb_parser.add_argument('--classic-notebook', action='store_true', default=False,
                           help="Use classic notebook instead of jupyterlab.")
    nb_parser.add_argument('--no-browser', action='store_true', default=False,
                           help=("Start the jupyter server to serve the notebook "
                                 "but don't open the notebook in the browser.\n"
                                 "Use this option to connect remotely from localhost to the machine running the kernel"))
    nb_parser.add_argument('--foreground', action='store_true', default=False,
        help="Run jupyter notebook in the foreground.")

    parser = argparse.ArgumentParser(epilog=get_epilog() if with_epilog else "",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-V', '--version', action='version', version=abilab.__version__)

    # Parser for commands that need to call spglib.
    spgopt_parser = argparse.ArgumentParser(add_help=False)
    spgopt_parser.add_argument('--symprec', default=1e-3, type=float,
        help="""\
symprec (float): Tolerance for symmetry finding. Defaults to 1e-3,
which is fairly strict and works well for properly refined structures with atoms in the proper symmetry coordinates.
For structures with slight deviations from their proper atomic positions (e.g., structures relaxed with electronic structure
codes), a looser tolerance of 0.1 (the value used in Materials Project) is often needed.""")
    spgopt_parser.add_argument('--angle-tolerance', default=5.0, type=float,
        help="angle_tolerance (float): Angle tolerance for symmetry finding. Default: 5.0")
    spgopt_parser.add_argument("--no-time-reversal", default=False, action="store_true", help="Don't use time-reversal.")
    spgopt_parser.add_argument("--site-symmetry", default=False, action="store_true",
                               help="Show site symmetries i.e. the point group operations that leave the site invariant.")

    # Parent parser for common options.
    copts_parser = argparse.ArgumentParser(add_help=False)
    copts_parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
        help='verbose, can be supplied multiple times to increase verbosity')
    copts_parser.add_argument('--loglevel', default="ERROR", type=str,
        help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")

    # Parent parser for commands that need to save structure to file.
    savefile_parser = argparse.ArgumentParser(add_help=False)
    savefile_parser.add_argument("-s", "--savefile", default="", type=str,
        help="Save final structure to file. Format is detected from file extensions "
             "e.g. out.abi for Abinit input, out.cif for CIF format.")

    # Helper functions to construct sub-parsers.
    def add_primitive_options(parser):
        """Add --no-primitive and --primitive-standard options to a parser."""
        group = parser.add_mutually_exclusive_group()
        group.add_argument('--no-primitive', default=False, action='store_true', help="Do not enforce primitive cell.")
        group.add_argument('--primitive-standard', default=False, action='store_true',
                           help="Enforce primitive standard cell.")

    supported_formats = "(abivars, cif, xsf, poscar, qe, siesta, wannier90, cssr, json, None)"

    def add_format_arg(parser, default, option=True, formats=None):
        """
        Add --format option to a parser with default value `default`.
        """
        formats = supported_formats if formats is None else formats
        if option:
            parser.add_argument("-f", "--format", default=default, type=str,
                help="Output format. Default: %s. Accept: %s" % (default, formats))
        else:
            parser.add_argument('format', nargs="?", default=default, type=str,
                help="Output format. Default: %s. Accept: %s" % (default, formats))

    # Create the parsers for the sub-commands
    subparsers = parser.add_subparsers(dest='command', help='sub-command help',
        description="Valid subcommands, use command --help for help")

    # Subparser for spglib command.
    p_spglib = subparsers.add_parser('spglib', parents=[copts_parser, path_selector, spgopt_parser],
        help="Analyze structure with spglib.")

    # Subparser for abispg command.
    p_abispg = subparsers.add_parser('abispg', parents=[copts_parser, path_selector, savefile_parser],
        help="Extract/Compute Abinit space group from file with structure.")
    p_abispg.add_argument("-t", "--tolsym", type=float, default=None, help="""\
Gives the tolerance on the atomic positions (reduced coordinates), primitive vectors, or magnetization,
to be considered equivalent, thanks to symmetry operations. This value is used by ABINIT in the recognition of the set
of symmetries of the system, or the application of the symmetry operations to generate from a reduced set of atoms,
The internal default is 1e-8. Setting tolsym to a value larger than 1e-8 will make Abinit detect the spacegroup within
this tolerance and re-symmetrize the input structure. This option is useful if the structure has been taken from a CIF
file that does not have enough significant digits.""")
    p_abispg.add_argument("-d", "--diff-mode", type=str, default="table", choices=["table", "diff"],
        help="Select diff output format.")

    # Subparser for convert command.
    p_convert = subparsers.add_parser('convert', parents=[copts_parser, path_selector],
        help="Convert structure to the specified format.")
    add_format_arg(p_convert, default="cif")

    # Subparser for print command.
    p_print = subparsers.add_parser('print', parents=[copts_parser, path_selector],
                                    help="Print Structure to terminal.")

    # Subparser for supercell command.
    p_supercell = subparsers.add_parser('supercell', parents=[copts_parser, path_selector],
        help="Generate supercell.")
    p_supercell.add_argument("-s", "--scaling_matrix", nargs="+", required=True, type=int,
        help="""\
scaling_matrix: A scaling matrix for transforming the lattice vectors.
Has to be all integers. Several options are possible:

    a. A full 3x3 scaling matrix defining the linear combination
       the old lattice vectors. E.g., -s 2,1,0 0,3,0, 0,0,1 generates a new structure with lattice vectors
       a' = 2a + b, b' = 3b, c' = c where a, b, and c are the lattice vectors of the original structure.
    b. An sequence of three scaling factors. E.g., -s 2, 1, 1 specifies that the supercell should
       have dimensions 2a x b x c.
    c. A number, which simply scales all lattice vectors by the same factor.
    """)
    add_format_arg(p_supercell, default="abivars")

    # Subparser for abisanitize
    p_abisanitize = subparsers.add_parser('abisanitize', parents=[copts_parser, path_selector, spgopt_parser, savefile_parser],
        help="Sanitize structure with abi_sanitize, compare structures and save result to file.")
    add_primitive_options(p_abisanitize)

    p_primitive = subparsers.add_parser('primitive', parents=[copts_parser, path_selector, spgopt_parser, savefile_parser],
        help="Use spglib to find a smaller unit cell than the input")

    # Subparser for irefine
    p_irefine = subparsers.add_parser('irefine', parents=[copts_parser, path_selector, spgopt_parser],
        help="Refine structure with abi_sanitize iteratively, stop if target space group is obtained.")
    p_irefine.add_argument("--target-spgnum", required=True, type=int, help="Target space group number.")
    p_irefine.add_argument("--symprec-step", default=0.05, type=float, help='Increment for symprec.')
    p_irefine.add_argument("--angle-tolerance-step", default=0.0, type=float, help='Increment for angle_tolerance.')
    p_irefine.add_argument("--ntrial", default=50, type=int, help='Number of trials. Default 50.')
    add_primitive_options(p_irefine)

    # Subparser for conventional.
    p_conventional = subparsers.add_parser('conventional', parents=[copts_parser, path_selector, spgopt_parser, savefile_parser],
        help="Gives a structure with a conventional cell according to certain standards. "
             "The standards are defined in doi:10.1016/j.commatsci.2010.05.010")

    # Subparser for proto.
    p_proto = subparsers.add_parser('proto', parents=[copts_parser, path_selector],
        help=("Find prototype in the AFLOW LIBRARY OF CRYSTALLOGRAPHIC PROTOTYPES. "
              "http://doi.org/10.1016/j.commatsci.2017.01.017"))
    p_proto.add_argument("--ltol", default=0.2, type=float, help="fractional length tolerance.")
    p_proto.add_argument("--stol", default=0.3, type=float, help="site tolerance.")
    p_proto.add_argument("--angle-tol", default=5, type=float, help="angle tolerance.")

    # Subparser for wyckoff.
    p_wyckoff = subparsers.add_parser('wyckoff', parents=[copts_parser, spgopt_parser, path_selector],
            help="Print wyckoff positions. WARNING: still under development!")
    p_wyckoff.add_argument("--refine", default=False, action="store_true",
                           help="Use spglib to refine structure before computation")

    # Subparser for tensor_site.
    p_tensor_site = subparsers.add_parser('tensor_site', parents=[copts_parser, spgopt_parser, path_selector],
            help="Print symmetry properties of tensors due to site-symmetries. WARNING: still under development!")
    p_tensor_site.add_argument("--refine", default=False, action="store_true",
                               help="Use spglib to refine structure before computation")

    # Subparser for neighbors.
    p_neighbors = subparsers.add_parser('neighbors', parents=[copts_parser, path_selector],
                                        help="Get neighbors for each atom in the unit cell, out to a distance radius.")
    p_neighbors.add_argument("-r", "--radius", default=2, type=float, help="Radius of the sphere in Angstrom.")

    # Subparser for interpolate.
    p_interpolate = subparsers.add_parser('interpolate', parents=[copts_parser],
        help=("Interpolate between two structures. Useful for the construction of NEB inputs."))
    p_interpolate.add_argument("filepaths", nargs=2, help="Files with initial and final structures.")
    p_interpolate.add_argument("-n", "--nimages", default=10, type=int, help="No. of interpolation images. Defaults to 10.")
    p_interpolate.add_argument("--autosort_tol", default=0.5, type=float, help="""\
A distance tolerance in Angstrom in which to automatically sort end_structure to match to the
closest points in this particular structure. This is usually what you want in a NEB calculation.
0 implies no sorting. Otherwise, a 0.5 value (default) usually works pretty well.""")
    add_format_arg(p_interpolate, default="abivars")

    # Subparser for xrd.
    p_xrd = subparsers.add_parser('xrd', parents=[copts_parser, path_selector], help="X-ray diffraction plot.")
    p_xrd.add_argument("-w", "--wavelength", default="CuKa", type=str, help=(
        "The wavelength can be specified as a string. It must be one of the "
        "supported definitions in the WAVELENGTHS dict declared in pymatgen/analysis/diffraction/xrd.py."
        "Defaults to 'CuKa', i.e, Cu K_alpha radiation."))
    p_xrd.add_argument("-s", "--symprec", default=0, type=float, help=(
        "Symmetry precision for structure refinement. "
        "If set to 0, no refinement is done. Otherwise, refinement is performed using spglib with provided precision."))
    p_xrd.add_argument("-t", "--two-theta-range", default=(0, 90), nargs=2, help=(
        "Tuple for range of two_thetas to calculate in degrees. Defaults to (0, 90)."))
    p_xrd.add_argument("-nap", "--no-annotate-peaks", default=False, action="store_true",
        help="Whether to annotate the peaks with plane information.")

    # Subparser for oxistate.
    p_oxistate = subparsers.add_parser('oxistate', parents=[copts_parser, path_selector],
        help="Estimate oxidation states with pymatgen bond valence analysis.")
    # Subparser for ipython.
    p_ipython = subparsers.add_parser('ipython', parents=[copts_parser, path_selector],
        help="Open IPython shell for advanced operations on structure object.")
    # Subparser for notebook.
    p_notebook = subparsers.add_parser('notebook', parents=[copts_parser, path_selector],
        help="Read structure from file and generate jupyter notebook.")
    p_notebook.add_argument('--classic-notebook', action='store_true', default=False,
                            help="Use classic notebook instead of jupyterlab.")
    p_notebook.add_argument('--no-browser', action='store_true', default=False,
                            help=("Start the jupyter server to serve the notebook "
                                  "but don't open the notebook in the browser.\n"
                                  "Use this option to connect remotely from localhost to the machine running the kernel"))
    p_notebook.add_argument('--foreground', action='store_true', default=False,
        help="Run jupyter notebook in the foreground.")

    p_panel = subparsers.add_parser('panel', parents=[copts_parser, path_selector],
        help="Open GUI in web browser, requires panel package.")
    p_panel.add_argument("-pnt", "--panel-template", default="FastList", type=str,
                        help="Specify template for panel dasboard." +
                             "Possible values are: FastList, FastGrid, Golden, Bootstrap, Material, React, Vanilla." +
                             "Default: FastList"
                        )
    p_panel.add_argument("--port", default=0, type=int, help="Allows specifying a specific port when serving panel app.")
    p_panel.add_argument('--no-browser', action='store_true', default=False,
                           help=("Start the bokeh server to serve the panel app "
                                 "but don't open the notebook in the browser.\n"
                                 "Use this option to connect remotely from localhost to the machine running the kernel"))

    # Subparser for kpath.
    p_kpath = subparsers.add_parser('kpath', parents=[copts_parser, path_selector],
        help="Read structure from file, generate k-path for band-structure calculations.")
    add_format_arg(p_kpath, default="abinit", formats=["abinit", "wannier90", "siesta"])

    # Subparser for bz.
    p_bz = subparsers.add_parser('bz', parents=[copts_parser, path_selector],
        help="Read structure from file, plot Brillouin zone with matplotlib.")
    # Subparser for ngkpt.
    p_ngkpt = subparsers.add_parser('ngkpt', parents=[copts_parser, path_selector],
        help="Return the Abinit k-point sampling variables "
             "from the number of divisions used to sample the smallest "
             "lattice vector of the reciprocal lattice.")
    p_ngkpt.add_argument("-n", "--nksmall", required=True, type=int,
        help="Number of divisions used to sample the smallest reciprocal lattice vector.")
    # Subparser for ktables.
    p_ktables = subparsers.add_parser('ktables', parents=[copts_parser, path_selector],
        help=("Read structure from filepath, call spglib to sample the BZ, "
              "print k-points in the IBZ with weights."))
    p_ktables.add_argument("-m", "--mesh", nargs=3, required=True, type=int, help="Mesh divisions e.g. 2 3 4")
    p_ktables.add_argument("-s", "--is_shift", nargs="+", default=None, type=int,
        help=("Three integers (spglib API). The kmesh is shifted along " +
              "the axis in half of adjacent mesh points irrespective of the mesh numbers e.g. 1 1 1 " +
              "Default: Unshifted mesh."))
    p_ktables.add_argument("--no-time-reversal", default=False, action="store_true", help="Don't use time-reversal.")

    # Subparser for abikmesh.
    p_abikmesh = subparsers.add_parser('abikmesh', parents=[copts_parser, path_selector],
        help=("Read structure from file, call Abinit to sample the BZ with ngkpt, shiftk, and kptopt. "
              "Print k-points in the IBZ with weights."))
    p_abikmesh.add_argument("--kppa", default=None, type=int,
            help="Number of k-points per reciprocal atom. Mutually exclusive with ngkpt.")
    p_abikmesh.add_argument("--ngkpt", nargs=3, default=None, type=int,
            help="Mesh divisions e.g. 2 3 4")
    p_abikmesh.add_argument("--shiftk", nargs="+", default=(0.5, 0.5, 0.5), type=float,
       help="Kmesh shifts. Default: 0.5 0.5 0.5")
    p_abikmesh.add_argument("--kptopt", default=1, type=int,
            help="Kptopt input variable. Default: 1")

    # Subparser for kmesh_jhu.
    #p_kmesh_jhu = subparsers.add_parser('kmesh_jhu', parents=[copts_parser, path_selector], help="Foobar ")

    # Subparser for lgk.
    p_lgk = subparsers.add_parser('lgk', parents=[copts_parser, path_selector, spgopt_parser],
        help="Read structure from file, find little group of k-point, print Bilbao character table.")
    p_lgk.add_argument("-k", "--kpoint", nargs=3, required=True, type=float,
        help="K-point in reduced coordinates e.g. 0.25 0 0")

    # Subparser for kstar.
    p_kstar = subparsers.add_parser('kstar', parents=[copts_parser, path_selector, spgopt_parser],
        help="Read structure from file, print star of k-point.")
    p_kstar.add_argument("-k", "--kpoint", nargs=3, required=True, type=float,
        help="K-point in reduced coordinates e.g. 0.25 0 0")

    # Subparser for keq.
    p_keq = subparsers.add_parser('keq', parents=[copts_parser, path_selector, spgopt_parser],
        help="Read structure from file, check whether two k-points are equivalent by symmetry.")
    p_keq.add_argument("-k", "--kpoints", nargs=6, required=True, type=float,
        help="K-points in reduced coordinates e.g. 0.25 0 0 0 0.25 0")

    # Subparser for visualize command.
    p_visualize = subparsers.add_parser('visualize', parents=[copts_parser, path_selector],
        help=("Visualize the structure with the specified application. "
              "Requires external app or optional python modules (mayavi, vtk)."))
    p_visualize.add_argument("-a", "--appname", type=str, default="vesta",
        help=("Application name. Possible options: %s, mpl (matplotlib), mayavi, vtk" % ", ".join(Visualizer.all_visunames())))

    # Options for commands accessing the materials project database.
    mp_rest_parser = argparse.ArgumentParser(add_help=False)
    mp_rest_parser.add_argument("--mapi-key", default=None,
        help="Pymatgen PMG_MAPI_KEY. Use value in .pmgrc.yaml if not specified.")
    mp_rest_parser.add_argument("--endpoint", help="Pymatgen database.", default="https://www.materialsproject.org/rest/v2")
    mp_rest_parser.add_argument("-b", "--browser", default=False, action='store_true',
        help="Open materials-project webpages in browser")

    # Subparser for mp_id command.
    p_mpid = subparsers.add_parser('mp_id', parents=[copts_parser, mp_rest_parser],
        help="Get structure from the pymatgen database. Export to format. Requires internet connection and PMG_MAPI_KEY.")
    p_mpid.add_argument("mpid", type=str, default=None, help="Pymatgen identifier.")
    add_format_arg(p_mpid, default="cif")

    # Subparser for mp_match command.
    p_mpmatch = subparsers.add_parser('mp_match', parents=[path_selector, mp_rest_parser, copts_parser, nb_parser],
        help="Get structure from the pymatgen database. Requires internet connection and PMG_MAPI_KEY.")
    add_format_arg(p_mpmatch, default="abivars")

    # Subparser for mp_search command.
    p_mpsearch = subparsers.add_parser('mp_search', parents=[mp_rest_parser, copts_parser, nb_parser],
        help="Get structure from the pymatgen database. Requires internet connection and PMG_MAPI_KEY")
    p_mpsearch.add_argument("chemsys_formula_id", type=str, default=None,
        help="A chemical system (e.g., Li-Fe-O), or formula (e.g., Fe2O3) or materials_id (e.g., mp-149).")
    p_mpsearch.add_argument("-s", "--select-spgnum", type=int, default=None,
        help="Select structures with this space group number.")
    add_format_arg(p_mpsearch, default="abivars")

    # Subparser for mp_pd command.
    p_mp_pda = subparsers.add_parser('mp_pd', parents=[mp_rest_parser, copts_parser],
        help=("Generate phase diagram with entries from the Materials Project. "
              "Requires internet connection and PMG_MAPI_KEY"))
    p_mp_pda.add_argument("file_or_elements", type=str, default=None,
        help="FILE with structure or elements e.g., Li-Fe-O).")
    p_mp_pda.add_argument("-u", "--show-unstable", type=int, default=0,
        help="""Whether unstable phases will be plotted as
well as red crosses. If a number > 0 is entered, all phases with
ehull < show_unstable will be shown.""")

    # Subparser for mp_ebands command.
    p_mp_ebands = subparsers.add_parser('mp_ebands', parents=[copts_parser, mp_rest_parser],
        help="Get structure from the pymatgen database. Export to format. Requires internet connection and PMG_MAPI_KEY.")
    p_mp_ebands.add_argument("chemsys_formula_id", type=str, default=None,
        help="A chemical system (e.g., Li-Fe-O), or formula (e.g., Fe2O3) or materials_id (e.g., mp-149).")
    #add_format_arg(p_mp_ebands, default="cif")

    # Subparser for cod_search command.
    p_codsearch = subparsers.add_parser('cod_search', parents=[copts_parser, nb_parser],
        help="Get structure from COD database. Requires internet connection and mysql")
    p_codsearch.add_argument("formula", type=str, default=None, help="formula (e.g., Fe2O3).")
    p_codsearch.add_argument("-s", "--select-spgnum", type=int, default=None,
        help="Select structures with this space group number.")
    p_codsearch.add_argument('--primitive', default=False, action='store_true',
        help="Convert COD cells into primitive cells.")
    add_format_arg(p_codsearch, default="abivars")

    # Subparser for cod_id command.
    p_codid = subparsers.add_parser('cod_id', parents=[copts_parser],
        help="Get structure from COD database. Requires internet connection and mysql")
    p_codid.add_argument("cod_identifier", type=int, default=None, help="COD identifier.")
    p_codid.add_argument('--primitive', default=False, action='store_true', help="Convert COD cell into primitive cell.")
    add_format_arg(p_codid, default="abivars")

    # Subparser for animate command.
    p_animate = subparsers.add_parser('animate', parents=[copts_parser, path_selector],
        help="Read structures from HIST.nc or XDATCAR. Print structures in Xcrysden AXSF format to stdout.")

    # Subparser for chemenv command.
    p_chemenv = subparsers.add_parser('chemenv', parents=[copts_parser, path_selector],
        help="Use ChemEnv to analyze chemical coordination environments. " +
             "See D. Waroquiers, at al. Acta Cryst B 2020, 76, 683–695.")
    p_chemenv.add_argument("-mdf", '--maximum-distance-factor', default=1.41,
                           type=float, help="Maximum distance factor. Default 1.41.")
    p_chemenv.add_argument("-dc", '--distance-cutoff', default=1.4, type=float,
                           help="Distance cutoff in Ang. Default: 1.4")
    p_chemenv.add_argument("-ac", '--angle-cutoff', default=0.3, type=float,
                           help="Angle cutoff. Default: 0.3")
    p_chemenv.add_argument('-mw', '--multi-weights', default=False, action="store_true",
                           help='Use MultiWeightsChemenvStrategy instead of SimplestChemenvStrategy.')
    p_chemenv.add_argument("-i", "--only-indices", nargs="+", default=None, type=int,
                           help="List of site indices to analyze e.g. `-i 0 2`. Default: None i.e. all sites.")
    p_chemenv.add_argument("-s", "--only-atoms", nargs="+", default=None, type=str,
                           help="List of atomic symbols to analyze e.g. `-s Si O`. Default: None i.e. all sites.")
    return parser


def serve_kwargs_from_options(options) -> dict:
    if options.no_browser:
        print("""
Use:

    ssh -N -f -L localhost:{port}:localhost:{port} username@your_remote_cluster

for port forwarding.
""")

    import abipy.panels as mod
    assets_path = os.path.join(os.path.dirname(mod.__file__), "assets")

    return dict(
        debug=options.verbose > 0,
        show=not options.no_browser,
        port=options.port,
        static_dirs={"/assets": assets_path},
        #address=address,
        #websocket_origin="{address}:{port}",
    )


@prof_main
def main():

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg: sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    # Parse command line.
    try:
        options = parser.parse_args()
    except Exception as exc:
        show_examples_and_exit(error_code=1)

    if not options.command:
        show_examples_and_exit(error_code=1)

    if hasattr(options, "format"):
        options.format = options.format.strip()

    cli.set_loglevel(options.loglevel)

    if options.verbose > 2:
        print(options)

    if options.command == "spglib":
        structure = abilab.Structure.from_file(options.filepath)
        print(structure.spget_summary(symprec=options.symprec, angle_tolerance=options.angle_tolerance,
                                      site_symmetry=options.site_symmetry, verbose=options.verbose))

    elif options.command == "abispg":
        structure = abilab.Structure.from_file(options.filepath)
        check_ordered_structure(structure)
        abi_spg = structure.abi_spacegroup

        if abi_spg is not None and options.tolsym is None:
            print(structure.spget_summary(verbose=options.verbose))
        else:
            # Here we compare Abinit wrt spglib. If abi_spg is None, we create a temporary
            # task to run the code in dry-run mode.
            if abi_spg is None:
                print("FILE does not contain Abinit symmetry operations.")
            cprint("Calling Abinit in --dry-run mode with chkprim = 0 to get space group.")
            if options.tolsym is not None and options.tolsym > 1e-8:
                cprint("Crystal structure will be re-symmetrized by Abinit with tolsym: %s" % options.tolsym, "yellow")

            from abipy.data.hgh_pseudos import HGH_TABLE
            gsinp = factories.gs_input(structure, HGH_TABLE, spin_mode="unpolarized")
            gsinp["chkprim"] = 0
            abistructure = gsinp.abiget_spacegroup(tolsym=options.tolsym)
            print(abistructure.spget_summary(verbose=options.verbose))
            print("")

            diff_structures([structure, abistructure], mode=options.diff_mode,
                            headers=["Input structure", "After Abinit symmetrization"], fmt="abivars")

            # Save file.
            save_structure(abistructure, options)

    elif options.command == "convert":
        fmt = options.format
        if fmt == "cif" and options.filepath.endswith(".cif"): fmt = "abivars"
        print(abilab.Structure.from_file(options.filepath).convert(fmt=fmt))

    elif options.command == "print":
        print(abilab.Structure.from_file(options.filepath).to_string(verbose=options.verbose))

    elif options.command == "supercell":
        structure = abilab.Structure.from_file(options.filepath)

        options.scaling_matrix = np.array(options.scaling_matrix)
        if len(options.scaling_matrix) == 9:
            options.scaling_matrix.shape = (3, 3)
        if options.verbose:
            print("scaling matrix: ", options.scaling_matrix)

        supcell = structure * options.scaling_matrix
        #supcell = structure.make_supercell(scaling_matrix, to_unit_cell=True)
        print("nsym 1")
        print("chkprim 0")
        print(supcell.convert(fmt=options.format))

    elif options.command == "abisanitize":
        print("\nCalling abi_sanitize to get a new structure in which:")
        print("    * Structure is refined.")
        print("    * Reduced to primitive settings.")
        print("    * Lattice vectors are exchanged if the triple product is negative\n")

        structure = abilab.Structure.from_file(options.filepath)
        sanitized = structure.abi_sanitize(symprec=options.symprec, angle_tolerance=options.angle_tolerance,
                                           primitive=not options.no_primitive, primitive_standard=options.primitive_standard)
        index = [options.filepath, "abisanitized"]
        dfs = abilab.dataframes_from_structures([structure, sanitized], index=index, with_spglib=True)
        abilab.print_dataframe(dfs.lattice, title="Lattice parameters:")
        abilab.print_dataframe(dfs.coords, title="Atomic positions in frac. coords. (columns give the site index):")

        if not options.verbose:
            print("\nUse -v for more info")
        else:
            #print("\nDifference between structures:")
            if len(structure) == len(sanitized):
                table = []
                for line1, line2 in zip(str(structure).splitlines(), str(sanitized).splitlines()):
                    table.append([line1, line2])
                print(str(tabulate(table, headers=["Initial structure", "Abisanitized"])))

            else:
                print("\nInitial structure:")
                print(structure)
                print("\nabisanitized structure:")
                print(sanitized)

        # Save file.
        save_structure(sanitized, options)

    elif options.command == "primitive":
        structure = abilab.Structure.from_file(options.filepath)
        primitive = structure.get_primitive_structure(tolerance=0.25, use_site_props=False, constrain_latt=None)
        # Cast to abipy structure because get_primitive may break inheritance.
        primitive = abilab.Structure.as_structure(primitive)
        separator = "\n" + 90 * "="
        print("\nInitial structure:\n", structure, separator)
        print("\nPrimitive structure returned by pymatgen:\n", primitive, separator)
        print("\nAbinit input for primitive structure:\n", primitive.abi_string)
        if structure != primitive:
            print("\nInput structure is not primitive (according to pymatgen + spglib)")
        else:
            print("\nInput structure seems to be primitive (according to pymatgen + spglib)")

    elif options.command == "irefine":
        structure = abilab.Structure.from_file(options.filepath)
        sanitized = structure.copy()
        symprec, angle_tolerance = options.symprec, options.angle_tolerance
        print("Calling abi_sanitize with increasing tolerances to reach target space group:", options.target_spgnum)
        print("Using symprec_step: ", options.symprec_step, ", angle_tolerance_step:", options.angle_tolerance_step,
              "ntrial", options.ntrial)
        itrial = 0
        while itrial < options.ntrial:
            print(">>> Trying with symprec: %s, angle_tolerance: %s" % (symprec, angle_tolerance))
            sanitized = sanitized.abi_sanitize(symprec=symprec, angle_tolerance=angle_tolerance,
                primitive=not options.no_primitive, primitive_standard=options.primitive_standard)
            spg_symb, spg_num = sanitized.get_space_group_info(symprec=symprec, angle_tolerance=angle_tolerance)
            print(">>> Space-group number:", spg_symb, ", symbol:", spg_num, "for trial:", itrial)
            if spg_num == options.target_spgnum:
                print(2 * "\n", "# Final structure with space group number:", spg_symb, ", symbol:", spg_num, 2 * "\n")
                print(sanitized.convert(fmt="cif"))
                break

            # Increment counter and tolerances.
            itrial += 1
            symprec += options.symprec_step
            angle_tolerance += options.angle_tolerance_step
        else:
            print("Cannot find space group number:", options.target_spgnum, "after", options.ntrial, "iterations")
            return 1

        # Save file.
        #save_structure(sanitized, options)

    elif options.command == "conventional":
        print("\nCalling get_conventional_standard_structure to get conventional structure:")
        print("The standards are defined in Setyawan, W., & Curtarolo, S. (2010).")
        print("High-throughput electronic band structure calculations: Challenges and tools. ")
        print("Computational Materials Science, 49(2), 299-312. doi:10.1016/j.commatsci.2010.05.010\n")

        structure = abilab.Structure.from_file(options.filepath)
        conv = structure.get_conventional_standard_structure(international_monoclinic=True,
                                           symprec=options.symprec, angle_tolerance=options.angle_tolerance)
        index = [options.filepath, "conventional"]
        dfs = abilab.dataframes_from_structures([structure, conv], index=index, with_spglib=True)

        abilab.print_dataframe(dfs.lattice, title="Lattice parameters:")
        if options.verbose:
            abilab.print_dataframe(dfs.coords, title="Atomic positions in frac. coords (columns give the site index):")

        if not options.verbose:
            print("\nUse -v for more info")
        else:
            #print("\nDifference between structures:")
            if len(structure) == len(conv):
                table = []
                for line1, line2 in zip(str(structure).splitlines(), str(conv).splitlines()):
                    table.append([line1, line2])
                print(str(tabulate(table, headers=["Initial structure", "Conventional"])))

            else:
                print("\nInitial structure:\n", structure)
                print("\nConventional structure:\n", conv)

        # Save file.
        save_structure(conv, options)

    elif options.command == "proto":
        structure = abilab.Structure.from_file(options.filepath)
        from pymatgen.analysis.aflow_prototypes import AflowPrototypeMatcher
        m = AflowPrototypeMatcher(initial_ltol=options.ltol, initial_stol=options.stol,
                                  initial_angle_tol=options.angle_tol)
        dlist = m.get_prototypes(structure)
        if not dlist:
            cprint("Cannot find AFLOW prototype for structure.")
            print(structure.to_string(verbose=options.verbose))
            return 1
        else:
            cprint("Found %d matches" % len(dlist), "green")
            for d in dlist:
                if "snl" in d:
                    snl = d.pop("snl")
                    if options.verbose: pprint(snl.as_dict())
                pprint(d)
                url = "http://aflow.org/CrystalDatabase/%s.html" % d["tags"]["aflow"]
                print("AFLOW url: %s\n" % url)
            if not options.verbose:
                print("Use --verbose to increase output level")

    elif options.command == "wyckoff":
        structure = abilab.Structure.from_file(options.filepath)
        if options.refine:
            print("Refining structure with symprec: %s, angle_tolerance: %s" % (options.symprec, options.angle_tolerance))
            structure = structure.refine(symprec=options.symprec, angle_tolerance=options.angle_tolerance)
        print(structure.spget_summary(verbose=options.verbose))
        ss = structure.site_symmetries
        df = ss.get_wyckoff_dataframe(verbose=options.verbose)
        abilab.print_dataframe(df, title="\nWyckoff positions in reduced coordinates.")

    elif options.command == "tensor_site":
        structure = abilab.Structure.from_file(options.filepath)
        if options.refine:
            print("Refining structure with symprec: %s, angle_tolerance: %s" % (options.symprec, options.angle_tolerance))
            structure = structure.refine(symprec=options.symprec, angle_tolerance=options.angle_tolerance)
        print(structure.spget_summary(verbose=options.verbose))
        ss = structure.site_symmetries
        df = ss.get_tensor_rank2_dataframe(verbose=options.verbose)
        abilab.print_dataframe(df, title="\nTensor components in reduced coordinates (rank 2, symmetric)")

    elif options.command == "neighbors":
        abilab.Structure.from_file(options.filepath).print_neighbors(radius=options.radius)

    elif options.command == "interpolate":
        initial_structure = abilab.Structure.from_file(options.filepaths[0])
        end_structure = abilab.Structure.from_file(options.filepaths[1])
        structures = initial_structure.interpolate(end_structure, nimages=options.nimages,
                                                   interpolate_lattices=False, pbc=True,
                                                   autosort_tol=options.autosort_tol)
        structures = list(map(abilab.Structure.as_structure, structures))
        for i, s in enumerate(structures):
            print(marquee("Structure #%d" % i, mark="="))
            print(s.convert(fmt=options.format))
            print(" ")

    elif options.command == "xrd":
        structure = abilab.Structure.from_file(options.filepath)
        two_theta_range = tuple(float(t) for t in options.two_theta_range)
        structure.plot_xrd(wavelength=options.wavelength, two_theta_range=two_theta_range,
                           symprec=options.symprec, annotate_peaks=not options.no_annotate_peaks)

    elif options.command == "oxistate":
        print(abilab.Structure.from_file(options.filepath).get_oxi_state_decorated())

    elif options.command == "ipython":
        structure = abilab.Structure.from_file(options.filepath)
        print("Invoking Ipython, `structure` object will be available in the Ipython terminal")
        import IPython
        IPython.start_ipython(argv=[], user_ns={"structure": structure})

    elif options.command == "notebook":
        structure = abilab.Structure.from_file(options.filepath)
        structure.make_and_open_notebook(nbpath=None, foreground=options.foreground,
                                         classic_notebook=options.classic_notebook,
                                         no_browser=options.no_browser)

    elif options.command == "panel":
        structure = abilab.Structure.from_file(options.filepath)
        pn = abilab.abipanel()
        serve_kwargs = serve_kwargs_from_options(options)
        app = structure.get_panel(template=options.panel_template)
        return pn.serve(app, **serve_kwargs)

    elif options.command == "visualize":
        structure = abilab.Structure.from_file(options.filepath)
        print(structure)
        print("Visualizing structure with:", options.appname)
        structure.visualize(appname=options.appname)

    elif options.command == "kpath":
        structure = abilab.Structure.from_file(options.filepath)
        print(structure.get_kpath_input_string(fmt=options.format, line_density=10))

    elif options.command == "bz":
        abilab.Structure.from_file(options.filepath).plot_bz()

    elif options.command == "ngkpt":
        d = abilab.Structure.from_file(options.filepath).calc_ksampling(options.nksmall)
        print("ngkpt %d %d %d" % (d.ngkpt[0], d.ngkpt[1], d.ngkpt[2]))
        print("nshiftk ", len(d.shiftk), "\nshiftk")
        for s in d.shiftk:
            print("  %s %s %s" % (s[0], s[1], s[2]))

    elif options.command == "ktables":
        structure = abilab.Structure.from_file(options.filepath)
        k = Ktables(structure, options.mesh, options.is_shift, not options.no_time_reversal)
        print(k)
        print("")
        print("NB: These results are obtained by calling spglib with the structure read from file.")
        print("The k-points might differ from the ones expected by Abinit, especially if the space groups differ.")

        if not options.verbose:
            print("\nUse -v to obtain the BZ --> IBZ mapping.")
        else:
            print()
            k.print_bz2ibz()

    elif options.command == "abikmesh":
        structure = abilab.Structure.from_file(options.filepath)
        if options.kppa is None and options.ngkpt is None:
            raise ValueError("Either ngkpt or kppa must be provided")

        if options.kppa is not None:
            print("Calling Abinit to compute the IBZ with kppa:", options.kppa, "and shiftk:", options.shiftk)
            ibz = IrredZone.from_kppa(structure, options.kppa, options.shiftk,
                                      kptopt=options.kptopt, verbose=options.verbose)
        else:
            print("Calling Abinit to compute the IBZ with ngkpt:", options.ngkpt, "and shiftk:", options.shiftk)
            ibz = IrredZone.from_ngkpt(structure, options.ngkpt, options.shiftk,
                                       kptopt=options.kptopt, verbose=options.verbose)

        print(ibz.to_string(verbose=options.verbose))

    #elif options.command == "kmesh_jhu":
    #    structure = abilab.Structure.from_file(options.filepath)
    #    from pymatgen.ext.jhu import get_kpoints
    #    kpoints = get_kpoints(structure, min_distance=0, min_total_kpoints=1,
    #                           kppra=None, gap_distance=7, remove_symmetry=None,
    #                           include_gamma="auto", header="simple", incar=None)
    #    #print(kpoints)

    elif options.command == "lgk":
        structure = abilab.Structure.from_file(options.filepath)
        spgrp = structure.abi_spacegroup
        if spgrp is None:
            cprint("Your file does not contain Abinit symmetry operations.", "yellow")
            cprint("Will call spglib to obtain the space group (assuming time-reversal: %s)" %
                   (not options.no_time_reversal), "yellow")
            spgrp = AbinitSpaceGroup.from_structure(structure, has_timerev=not options.no_time_reversal,
                        symprec=options.symprec, angle_tolerance=options.angle_tolerance)
        print()
        print(marquee("Structure", mark="="))
        print(structure.spget_summary(verbose=options.verbose))
        print("\n")

        print(marquee("Little Group", mark="="))
        ltk = spgrp.find_little_group(kpoint=options.kpoint)
        print(ltk.to_string(verbose=options.verbose))

    elif options.command == "kstar":
        structure = abilab.Structure.from_file(options.filepath)

        # Call spglib to get spacegroup if Abinit spacegroup is not available.
        if structure.abi_spacegroup is None:
            structure.spgset_abi_spacegroup(has_timerev=not options.no_time_reversal)

        kpoint = Kpoint(options.kpoint, structure.reciprocal_lattice)
        kstar = kpoint.compute_star(structure.abi_spacegroup, wrap_tows=True)
        print("Found %s points in the star of %s\n" % (len(kstar), repr(kpoint)))
        for k in kstar:
            print(4 * " ", repr(k))

    elif options.command == "keq":
        structure = abilab.Structure.from_file(options.filepath)

        # Call spglib to get spacegroup if Abinit spacegroup is not available.
        if structure.abi_spacegroup is None:
            structure.spgset_abi_spacegroup(has_timerev=not options.no_time_reversal)

        k1, k2 = options.kpoints[:3], options.kpoints[3:6]
        k1tab = structure.abi_spacegroup.symeq(k1, k2)

        if k1tab.isym != -1:
            print("\nk1:", k1, "and k2:", k2, "are symmetry equivalent k-points\n")
            print("Related by the symmetry operation (reduced coords):\n", k1tab.op)
            print("With umklapp vector Go = TO(k1) - k2 =", k1tab.g0)
        else:
            print(k1, "and", k2, "are NOT symmetry equivalent")

    elif options.command == "mp_id":
        # Get the Structure corresponding to material_id.
        structure = abilab.Structure.from_mpid(options.mpid, final=True,
                                               api_key=options.mapi_key, endpoint=options.endpoint)
        # Convert to format and print it.
        print(structure.convert(fmt=options.format))

    elif options.command == "mp_match":
        mp = abilab.mp_match_structure(options.filepath)
        if not mp.structures:
            cprint("No structure found in MP database", "yellow")
            return 1

        if options.notebook:
            return mp.make_and_open_notebook(foreground=options.foreground,
                                             classic_notebook=options.classic_notebook,
                                             no_browser=options.no_browser)
        else:
            mp.print_results(fmt=options.format, verbose=options.verbose)

        if options.browser:
            mp.open_browser(limit=None if options.verbose == 2 else 10)

    elif options.command == "mp_search":
        mp = abilab.mp_search(options.chemsys_formula_id)
        if not mp.structures:
            cprint("No structure found in Materials Project database", "yellow")
            return 1
        if options.select_spgnum: mp = mp.filter_by_spgnum(options.select_spgnum)

        if options.notebook:
            return mp.make_and_open_notebook(foreground=options.foreground,
                                             classic_notebook=options.classic_notebook,
                                             no_browser=options.no_browser)
        else:
            mp.print_results(fmt=options.format, verbose=options.verbose)

        if options.browser:
            mp.open_browser(limit=None if options.verbose == 2 else 10)

    elif options.command == "mp_pd":
        if os.path.exists(options.file_or_elements):
            structure = abilab.Structure.from_file(options.file_or_elements)
            elements = structure.symbol_set
        else:
            elements = options.file_or_elements.split("-")

        if options.verbose > 1: print("Building phase-diagram for elements:", elements)
        with abilab.restapi.get_mprester(api_key=options.mapi_key, endpoint=options.endpoint) as rest:
            pdr = rest.get_phasediagram_results(elements)
            pdr.print_dataframes(verbose=options.verbose)
            pdr.plot(show_unstable=options.show_unstable)

    elif options.command == "mp_ebands":
        if os.path.exists(options.chemsys_formula_id):
            mp = abilab.mp_match_structure(options.chemsys_formula_id)
            for mpid in mp.ids:
                ebands = abilab.ElectronBands.from_mpid(mpid, api_key=options.mapi_key, endpoint=options.endpoint)
                print(ebands)
        else:
            if options.chemsys_formula_id.startswith("mp-"):
                # Assume valid mp identifier.
                mpid = options.chemsys_formula_id
                ebands = abilab.ElectronBands.from_mpid(mpid, api_key=options.mapi_key, endpoint=options.endpoint)
                print(ebands)
            else:
                mp = abilab.mp_search(options.chemsys_formula_id)
                if not mp.structures:
                    cprint("No structure found in Materials Project database", "yellow")
                    return 1

                for mpid, structure in zip(mp.ids, mp.structures):
                    if structure is None:
                        cprint("ignoring mpid %s because cannot find structure" % mpid, "red")
                        continue
                    ebands = abilab.ElectronBands.from_mpid(mpid, api_key=options.mapi_key, endpoint=options.endpoint)
                    if ebands is None:
                        cprint("Cannot get ebands for structure:\n%s" % str(structure), "red")
                    else:
                        print(ebands)

    elif options.command == "cod_search":
        cod = abilab.cod_search(options.formula, primitive=options.primitive)
        if not cod.structures:
            cprint("No structure found in COD database", "yellow")
            return 1
        if options.select_spgnum: cod = cod.filter_by_spgnum(options.select_spgnum)

        if options.notebook:
            return cod.make_and_open_notebook(foreground=options.foreground,
                                              classic_notebook=options.classic_notebook,
                                              no_browser=options.no_browser)
        else:
            cod.print_results(fmt=options.format, verbose=options.verbose)

    elif options.command == "cod_id":
        # Get the Structure from COD
        structure = abilab.Structure.from_cod_id(options.cod_identifier, primitive=options.primitive)
        # Convert to format and print it.
        print(structure.convert(fmt=options.format))

    elif options.command == "animate":
        filepath = options.filepath
        if any(filepath.endswith(ext) for ext in ("HIST", "HIST.nc")):
            with abilab.abiopen(filepath) as hist:
                structures = hist.structures

        elif "XDATCAR" in filepath:
            structures = Xdatcar(filepath).structures
            if not structures:
                raise RuntimeError("Your Xdatcar contains only one structure. Due to a bug "
                    "in the pymatgen routine, your structures won't be parsed correctly"
                    "Solution: Add another structure at the end of the file.")
        else:
            raise ValueError("Don't know how to handle file %s" % filepath)

        xsf_write_structure(sys.stdout, structures)

    elif options.command == "chemenv":
        # Based on https://matgenb.materialsvirtuallab.org/2018/01/01/ChemEnv-How-to-automatically-identify-coordination-environments-in-a-structure.html

        from pymatgen.analysis.chemenv.coordination_environments.chemenv_strategies import SimplestChemenvStrategy
        from pymatgen.analysis.chemenv.coordination_environments.coordination_geometry_finder import LocalGeometryFinder
        from pymatgen.analysis.chemenv.coordination_environments.structure_environments import LightStructureEnvironments
        from pymatgen.analysis.chemenv.coordination_environments.chemenv_strategies import MultiWeightsChemenvStrategy

        import logging
        logging.basicConfig(
            format="%(levelname)s:%(module)s:%(funcName)s:%(message)s", level=logging.DEBUG
            # you can also save the logging to a file, just remove the comment
            # filename='chemenv_structure_environments.log',
        )

        structure = abilab.Structure.from_file(options.filepath)
        # Setup the local geometry finder
        lgf = LocalGeometryFinder()
        lgf.setup_structure(structure=structure)

        # Get the StructureEnvironments
        kws = dict(maximum_distance_factor=options.maximum_distance_factor,
                   only_indices=options.only_indices,
                   only_atoms=options.only_atoms,
                   #excluded_atoms=None,
                   #only_cations=True,
                   )
        print("Computing structure environments with:", kws)
        se = lgf.compute_structure_environments(**kws)

        if options.multi_weights:
            print("Using MultiWeightsChemenvStrategy with stats_article_weights_parameters")
            strategy = MultiWeightsChemenvStrategy.stats_article_weights_parameters()
        else:
            kws = dict(distance_cutoff=options.distance_cutoff, angle_cutoff=options.angle_cutoff)
            print("Using SimplestChemenvStrategy with:", kws)
            strategy = SimplestChemenvStrategy(**kws)

        lse = LightStructureEnvironments.from_structure_environments(
            strategy=strategy, structure_environments=se)

        lenv_keys = ["ce_symbol", "ce_fraction", "csm"]
        site_keys = ["site_index", "species", "xred1", "xred2", "xred3"]
        import pandas as pd

        if options.multi_weights:
            # One dataframe per site as we can have multiple csm per site.
            errored_isites = []
            for isite, site in enumerate(structure):
                if options.only_indices is not None and isite not in options.only_indices: continue
                d_list = lse.coordination_environments[isite]
                if d_list is None:
                    errored_isites.append(isite)
                    continue

                for i, d in enumerate(d_list):
                    d = {k: d[k] for k in lenv_keys}
                    d.update(list(zip(site_keys, [isite, site.species, *site.frac_coords])))
                    d_list[i] = d

                df = pd.DataFrame(d_list).set_index("site_index")
                print("")
                print(df.to_string())

        else:
            # One dataframe with all sites and one csm per site.
            d_list, errored_isites = [], []
            for isite, site in enumerate(structure):
                if options.only_indices is not None and isite not in options.only_indices: continue
                dl = lse.coordination_environments[isite]
                if dl is None:
                    errored_isites.append(isite)
                    continue
                d = dl[0]
                d = {k: d[k] for k in lenv_keys}
                d.update(list(zip(site_keys, [isite, site.species, *site.frac_coords])))
                d_list.append(d)

            df = pd.DataFrame(d_list).set_index("site_index")
            print("")
            print(df.to_string())

        if errored_isites:
            print("\nNB: Couldn't detect local env for", len(errored_isites), "site indices: ", errored_isites)

        print("\nFor the meaning of ce_symbol, see https://pubs.acs.org/doi/10.1021/acs.chemmater.7b02766")

    else:
        raise ValueError("Unsupported command: %s" % options.command)

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
ElectronBandsPlotter
====================

This example shows how to plot several band structures on a grid

We use two GSR files:

    si_scf_GSR.n: energies on a homogeneous sampling of the BZ (can be used to compute DOS)
    si_nscf_GSR.nc: energies on a k-path in the BZ (used to plot the band dispersion)
"""
from abipy.abilab import ElectronBandsPlotter
from abipy.data import ref_file

# To plot a grid with two band structures:
plotter = ElectronBandsPlotter()
plotter.add_ebands("BZ sampling", ref_file("si_scf_GSR.nc"))
plotter.add_ebands("k-path", ref_file("si_nscf_GSR.nc"))

# Get pandas dataframe
frame = plotter.get_ebands_frame()
print(frame)

#%%
# To create a grid plot use:
plotter.gridplot(with_gaps=True)

#%%
# Plotly version:
plotter.gridplotly(with_gaps=True)

#%%
# To plot a grid with band structures + DOS, use the optional argument `edos_objects`
# The first subplot gets the band dispersion from eb_objects[0] and the DOS from edos_objects[0]
# edos_kwargs is an optional dictionary passed to `get_dos` to compute the DOS.
#

eb_objects = 2 * [ref_file("si_nscf_GSR.nc")]
edos_objects = 2 * [ref_file("si_scf_GSR.nc")]

plotter = ElectronBandsPlotter()
plotter.add_ebands("Si", ref_file("si_nscf_GSR.nc"), edos=ref_file("si_scf_GSR.nc"))
plotter.add_ebands("Same data", ref_file("si_nscf_GSR.nc"), edos=ref_file("si_scf_GSR.nc"))
# sphinx_gallery_thumbnail_number = 2
plotter.gridplot()

#%%
# Plotly version:
plotter.gridplotly()



#!/usr/bin/env python
r"""
Flow to analyze the convergen of phonons in metals wrt ngkpt and tsmear
=======================================================================

This examples shows how to build a Flow to compute the
phonon band structure in a metallic system (MgB2) with different
k-point samplings and values of the electronic smearing tsmear
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import sys
import abipy.data as abidata
import abipy.abilab as abilab
from abipy import flowtk


def make_scf_input(structure, ngkpt, tsmear, pseudos, paral_kgb=1):
    """return GS input."""

    scf_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    # Global variables
    scf_inp.set_vars(
        ecut=35,
        nband=8,
        occopt=4,    # Marzari smearing
        tsmear=tsmear,
        paral_kgb=paral_kgb,
   )

    # Dataset 1 (GS run)
    scf_inp.set_kmesh(ngkpt=ngkpt, shiftk=structure.calc_shiftk())
    scf_inp.set_vars(tolvrs=1e-10)

    return scf_inp


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    structure = abidata.structure_from_ucell("MgB2")

    # Get pseudos from a table.
    table = abilab.PseudoTable(abidata.pseudos("12mg.pspnc", "5b.pspnc"))
    pseudos = table.get_pseudos_for_structure(structure)

    nval = structure.num_valence_electrons(pseudos)
    #print(nval)

    flow = flowtk.Flow(workdir=options.workdir)

    scf_work = flowtk.Work()
    ngkpt_list = [[4, 4, 4], [8, 8, 8], [12, 12, 12]]
    tsmear_list = [0.01, 0.02, 0.04]
    for ngkpt in ngkpt_list:
        for tsmear in tsmear_list:
            scf_input = make_scf_input(structure, ngkpt, tsmear, pseudos)
            scf_work.register_scf_task(scf_input)
    flow.register_work(scf_work)

    # This call uses the information reported in the GS task to
    # compute all the independent atomic perturbations corresponding to a [4, 4, 4] q-mesh.
    for scf_task in scf_work:
        ph_work = flowtk.PhononWork.from_scf_task(scf_task, qpoints=[4, 4, 4], is_ngqpt=True)
        flow.register_work(ph_work)

    return flow.allocate(use_smartio=True)


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).plot_networkx(with_edge_labels=True, tight_layout=True)


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_mgb2_phonons_nkpt_tsmear.py -s
#
# then use:
#
#    abirun.py flow_mgb2_phonons_nkpt_tsmear phbands
#
# to get info about the electronic properties:
#
# .. code-block:: shell
#
# Our main goal is to analyze the convergence of the phonons wrt to the k-point sampling and tsmear.
# As we know that ``w0_t2``, `w0_t3`` and ``w0_t4`` are DOS calculations, we can
# build a GSR robot for these tasks with:
#
#    abirun.py flow_mgb2_edoses/ ebands -nids=241034,241035,241036
#
# then, inside the ipython shell, one can use:
#
# .. code-block:: ipython
#
#    In [1]: %matplotlib
#    In [2]: robot.combiplot_edos()
#
# to plot the electronic DOS obtained with different number of k-points in the IBZ.
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_mgb2_edoses.png?raw=true
#    :alt: Convergence of electronic DOS in MgB2 wrt k-points.
#



# coding: utf-8
"""Wavefunction file."""
from __future__ import annotations

#import numpy as np

from monty.functools import lazy_property
from monty.string import marquee
from abipy.core import Mesh3D, GSphere
from abipy.core.structure import Structure
from abipy.core.mixins import AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.iotools import Visualizer
from abipy.electrons.ebands import ElectronsReader, ElectronBands
from abipy.waves.pwwave import PWWaveFunction
from abipy.tools import duck

__all__ = [
    "WfkFile",
]


class WfkFile(AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    This object provides a simplified interface to access and analyze
    the data stored in the WFK.nc file produced by ABINIT.

    Usage example:

    .. code-block:: python

        wfk = WfkFile("foo_WFK.nc")

        # Plot band energies.
        wfk.ebands.plot_ebands()

        # Visualize crystalline structure with vesta.
        wfk.visualize_structure_with("vesta")

        # Visualize u(r)**2 with vesta.
        wfk.visualize_ur2(spin=0, kpoint=0, band=0, appname="vesta")

        # Get a wavefunction.
        wave = wfk.get_wave(spin=0, kpoint=[0, 0, 0], band=0)

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: WfkFile
    """
    def __init__(self, filepath: str):
        """
        Initialize the object from a Netcdf file.
        """
        super().__init__(filepath)
        self.reader = reader = WFK_Reader(filepath)
        assert reader.has_pwbasis_set

        # Read the electron bands
        self._ebands = reader.read_ebands()

        self.npwarr = reader.npwarr
        self.nband_sk = reader.nband_sk

        # FFT mesh (augmented divisions reported in the WFK file)
        self.fft_mesh = Mesh3D(reader.fft_divs, self.structure.lattice_vectors())

        # Build G-spheres for each k-point
        gspheres = len(self.kpoints) * [None]
        ecut = reader.ecut
        for k, kpoint in enumerate(self.kpoints):
            gvec_k, istwfk = reader.read_gvecs_istwfk(k)
            gspheres[k] = GSphere(ecut, self.structure.reciprocal_lattice, kpoint, gvec_k, istwfk=istwfk)

        self._gspheres = tuple(gspheres)

        # Save reference to the reader.
        self.reader = reader

    def close(self) -> None:
        self.reader.close()

    @lazy_property
    def params(self) -> dict:
        """dict with parameters that might be subject to convergence studies."""
        od = self.get_ebands_params()
        return od

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.ebands.structure

    @property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| object"""
        return self._ebands

    @property
    def nkpt(self) -> int:
        """Number of k-points."""
        return len(self.kpoints)

    @property
    def gspheres(self) -> list:
        """List of :class:`GSphere` objects ordered by k-points."""
        return self._gspheres

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """
        String representation

        Args:
            verbose: verbosity level.
        """
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app(self.ebands.to_string(with_structure=False, verbose=verbose, title="Electronic Bands"))

        if verbose > 1:
            app("")
            app(self.hdr.to_string(verbose=verbose, title="Abinit Header"))

        return "\n".join(lines)

    def kindex(self, kpoint) -> int:
        """The index of the k-point in the file. Accepts :class:`Kpoint` object or int."""
        return self.reader.kindex(kpoint)

    def get_wave(self, spin, kpoint, band) -> PWWaveFunction:
        """
        Read and return the wavefunction with the given spin, band and kpoint.

        Args:
            spin: spin index. Must be in (0, 1)
            kpoint: Either :class:`Kpoint` instance or integer giving the sequential index in the IBZ (C-convention).
            band: band index.

            returns:
                :class:`WaveFunction` instance.
        """
        ik = self.kindex(kpoint)

        if (spin not in range(self.nsppol) or ik not in range(self.nkpt) or
            band not in range(self.nband_sk[spin, ik])):
            raise ValueError("Wrong (spin, band, kpt) indices")

        ug_skb = self.reader.read_ug(spin, kpoint, band)

        # Istantiate the wavefunction object and set the FFT mesh
        # using the divisions reported in the WFK file.
        wave = PWWaveFunction(self.structure, self.nspinor, spin, band, self.gspheres[ik], ug_skb)
        wave.set_mesh(self.fft_mesh)

        return wave

    def export_ur2(self, filepath, spin, kpoint, band, visu=None):
        """
        Export :math:`|u(r)|^2` on file filename.

        return:
            Instance of :class:`Visualizer`
        """
        # Read the wavefunction from file.
        wave = self.get_wave(spin, kpoint, band)

        # Export data uding the format specified by filename.
        if visu is None:
            return wave.export_ur2(filepath)
        else:
            return wave.export_ur2(filepath, visu=visu)

    def visualize_ur2(self, spin, kpoint, band, appname="vesta"):
        """
        Visualize :math:`|u(r)|^2`  with visualizer.
        See :class:`Visualizer` for the list of applications and formats supported.
        """
        visu = Visualizer.from_name(appname)

        for ext in visu.supported_extensions():
            ext = "." + ext
            try:
                v = self.export_ur2(ext, spin, kpoint, band, visu=visu)
                return v()
            except visu.Error:
                pass
        else:
            raise visu.Error("Don't know how to export data for visualizer %s" % appname)

    #def classify_states(self, spin, kpoint, band_range=None, energy_range=None, atol=1e-3):
    #    """
    #    Classify electronic eigenstates

    #    Args:
    #        spin: spin index. Must be in (0, 1)
    #        kpoint: Either :class:`Kpoint` instance or integer giving the sequential index in the IBZ (C-convention).
    #        band_range: Define the set of bands included in the classification. See also `energy_range`.
    #            `None` means all bands available in the WFK file. Accepts also range object, use e.g.
    #            `band_range=(0, 5)` to include [0, 1, 2, 3, 4].
    #        energy_range: Define the set of bands included in the classification. Mutually exclusive with `band_range`.
    #            Accepts: "gap"
    #        atol: Absolute tolerance on the energy in eV. Two states are considered degerate if
    #            their energy differ by less than `atol`.

    #    Return:
    #    """
    #    if band_range is not None and energy_range is not None:
    #        raise ValueError("band_range and energy_range are mutually exclusive.")

    #    if energy_range is None:
    #        bids = ... if band_range is None else list(range(band_range))
    #    else:
    #        raise NotImplementedError("energy_range")

    #    # Find little group of the k-point.
    #    ik = self.kindex(kpoint)
    #    kpoint = self.kpoints[ik]
    #    lgk = self.structure.abi_spacegroup.find_little_group(kpoint)
    #    #wclass = WavefuntionsClassifier(kpoint)

    #    # Select bands to include, find degenerate states and group them.
    #    from abipy.core.skw import find_degs_sk
    #    enes = self.ebands.eigens[spin, ik, bids]
    #    deg_list = find_degs_sk(enes, atol)

    #    for deg in deg_list:
    #        print("deg", deg)
    #        nb = len(deg)
    #        waves = [self.get_wave(spin, ik, band) for band in deg]
    #        mats = []
    #        for iclass, op_class in enumerate(lgk.groupby_class()):
    #            op = op_class[0]
    #            #print("op_class[0]\n", op)
    #            #op_waves = [wave.rotate(op) for wave in waves]
    #            op_waves = waves

    #            # Compute <u1|Op|u2>
    #            cmat = np.empty((nb, nb), dtype=np.complex)
    #            for i in range(nb):
    #                #print("int", waves[i].norm2())
    #                for j in range(nb):
    #                    cmat[i, j] = waves[i].braket(op_waves[j], space="r")
    #                    #cmat[i, j] = 1.0

    #            print(cmat)
    #            mats.append(cmat)

    #        #if wclass.classify_characters(deg, mats) /= 0:
    #        #    cprint("Warning")
    #        #    wclass.classify_characters_accidental(deg, full_mats)

    #    #return wclass

    def ipw_visualize_widget(self): # pragma: no cover
        """
        Return an ipython widget with controllers to visualize the wavefunctions.

        .. warning::

            It seems there's a bug with Vesta on MacOs if the user tries to open multiple wavefunctions
            as the tab in vesta is not updated!
        """
        def wfk_visualize(spin, kpoint, band, appname):
            kpoint = int(kpoint.split()[0])
            self.visualize_ur2(spin, kpoint, band, appname=appname)

        import ipywidgets as ipw
        return ipw.interact_manual(
                wfk_visualize,
                spin=list(range(self.nsppol)),
                kpoint=["%d %s" % (i, repr(kpt)) for i, kpt in enumerate(self.kpoints)],
                band=list(range(self.nband)),
                appname=[v.name for v in Visualizer.get_available()],
            )

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        yield self.ebands.plot(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write an ipython notebook to nbpath. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("wfk = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(wfk)"),
            nbv.new_code_cell("wfk.ebands.kpoints.plot();"),
            nbv.new_code_cell("wfk.ebands.plot();"),
            nbv.new_code_cell("""\
if wfk.ebands.kpoints.is_ibz:
    wfk.ebands.get_edos().plot();"""),
            nbv.new_code_cell("wfk.ipw_visualize_widget()"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class WFK_Reader(ElectronsReader):
    """
    This object reads data from the WFK file.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: Wfk_Reader
    """

    def __init__(self, filepath: str):
        """Initialize the object from a filename."""
        super().__init__(filepath)

        self.kpoints = self.read_kpoints()
        self.nfft1 = self.read_dimvalue("number_of_grid_points_vector1")
        self.nfft2 = self.read_dimvalue("number_of_grid_points_vector2")
        self.nfft3 = self.read_dimvalue("number_of_grid_points_vector3")

        self.cplex_ug = self.read_dimvalue("real_or_complex_coefficients")
        assert self.cplex_ug == 2

        self.nspinor = self.read_dimvalue("number_of_spinor_components")
        self.nsppol = self.read_dimvalue("number_of_spins")
        self.nspden = self.read_dimvalue("number_of_components")

        self.ecut = self.read_value("kinetic_energy_cutoff")
        self.nband_sk = self.read_value("number_of_states")
        self.istwfk = self.read_value("istwfk")
        self.npwarr = self.read_value("number_of_coefficients")

        # Store G-vectors
        self._kg = self.read_value("reduced_coordinates_of_plane_waves")

    @lazy_property
    def basis_set(self) -> str:
        """String defining the basis set."""
        basis_set = self.read_value("basis_set")
        return "".join(str(basis_set, encoding='UTF-8')).strip()

    @property
    def has_pwbasis_set(self) -> bool:
        """True if the plane-wave basis set is used."""
        return self.basis_set == "plane_waves"

    @property
    def fft_divs(self) -> tuple:
        """FFT divisions used to compute the data in the WFK file."""
        return self.nfft1, self.nfft2, self.nfft3

    def kindex(self, kpoint) -> int:
        """
        Index of the k-point in the internal tables.

        Accepts: :class:`Kpoint` instance or integer.
        """
        if duck.is_intlike(kpoint):
            return int(kpoint)
        else:
            return self.kpoints.index(kpoint)

    def read_gvecs_istwfk(self, kpoint) -> tuple:
        """
        Read the set of G-vectors and the value of istwfk for the given k-point.
        Accepts :class:`Kpoint` object or integer.
        """
        ik = self.kindex(kpoint)
        npw_k, istwfk = self.npwarr[ik], self.istwfk[ik]
        return self._kg[ik, :npw_k, :], istwfk

    def read_ug(self, spin, kpoint, band):
        """Read the Fourier components of the wavefunction."""
        ik = self.kindex(kpoint)
        npw_k, istwfk = self.npwarr[ik], self.istwfk[ik]
        if self.cplex_ug != 2:
            raise NotImplementedError("")

        # Read data from file (we don't store the full block full block in memory!).
        var = self.rootgrp.variables["coefficients_of_wavefunctions"]
        value = var[spin, ik, band, :, :npw_k, :]
        return value[..., 0] + 1j*value[..., 1]  # Build complex array



"""Panel objects to operate on DDB files."""
from __future__ import annotations

import sys
import param
import panel as pn
import panel.widgets as pnw

from abipy.core.structure import Structure
from abipy.panels.core import (AbipyParameterized, PanelWithStructure, BaseRobotPanel,
        mpl, ply, dfc, depends_on_btn_click, Loading, ActiveBar)
from abipy.dfpt.ddb import PhononBandsPlotter, DdbFile, DdbRobot
from abipy.tools.decorators import Appender


add_lr_docstring = Appender("""
The variables [[dipdip@anaddb]], [[dipquad@anaddb]] and [[quadquad@anaddb]]
define the treatment of the long-range part (LR) of the IFCs in the Fourier interpolation
of the dynamical matrix in semiconductors.
See XXX

By default, anaddb includes all the three different LR terms (DD, DQ, QQ)
to improve the quality of the Fourier interpolation **under the assumption** that
the DDB file contains the Born effective charges, the macroscopic dielectric tensor
and the dynamical quadrupoles.

This means that, in the majority of the cases, there is no need to change the default value
of [[dipdip@anaddb]], [[dipquad@anaddb]] and [[quadquad@anaddb]] unless
you want to disable them for testing purposes.
Note, however, that the fact these options are activated by default does not necessarily mean
that anaddb can take advantage of them.
Please look at the output in the **Summary** Tab to understand if your DDB file
contains the required entries.

""", indents=0)


class PanelWithAnaddbParams(param.Parameterized):
    """
    Base class providing widgets to invoke Anaddb via AbiPy.
    Used, for instance, by DdbFilePanel and DdbRobotPanel so that we don't have to
    repeat the same parameters over and over again.
    """

    nqsmall = param.Integer(10, bounds=(1, None), label="nqsmall (q-mesh for BZ integration)",
                           doc="Number of divisions for smallest vector to generate the q-mesh")

    ndivsm = param.Integer(5, bounds=(None, None), label="ndivsm (density of q-path)",
                          doc="Number of divisions for smallest vector to generate the q-path")

    lo_to_splitting = param.ObjectSelector(default="automatic", objects=["automatic", True, False])

    chneut = param.ObjectSelector(default=1, objects=[0, 1, 2], doc="Abinit variable")

    dipdip = param.ObjectSelector(default=1, objects=[0, 1, -1], doc="Abinit variable")

    dipquad = param.ObjectSelector(default=1, objects=[0, 1], doc="Abinit variable")

    quadquad = param.ObjectSelector(default=1, objects=[0, 1], doc="Abinit variable")

    asr = param.ObjectSelector(default=2, objects=[0, 1, 2], doc="Abinit variable")

    units = param.ObjectSelector(default="eV", objects=["eV", "meV", "Ha", "cm-1", "Thz"], doc="Energy units")

    dos_method = param.ObjectSelector(default="tetra", objects=["tetra", "gaussian"], label="Integration method for PDOS")
    #temp_range = param.Range(default=(200.0, 400.0), bounds=(0, 1000), label="Temperature range in K.")


    eps0_gamma_ev = param.Number(1e-4, bounds=(1e-20, None), label="Phonon linewidth in eV")
    #eps0_wrange = param.Range(default=(0.0, 0.1), bounds=(0.0, 1.0), label="Frequency range (eV)")

    ifc_yscale = param.ObjectSelector(default="log", label="Scale for IFCs",
                                      objects=["log", "linear", "symlog", "logit"])

    def __init__(self, **params):
        super().__init__(**params)
        # FIXME
        self.nqsmall_list = pnw.LiteralInput(name='nqmall_list (python syntax)', value=[10, 20, 30], type=list)
        #nqqpt = pnw.LiteralInput(name='nsmalls (list)', value=[10, 20, 30], type=list)

        self.eps0_wrange = pnw.EditableRangeSlider(name="Frequency range (eV)", value=(0.0, 0.1),
                                                   start=0.0, end=1.0, step=0.01)

        self.temp_range = pnw.EditableRangeSlider(name='T-range in K', value=(100.0, 400),
                                                  start=0, end=1000, step=50)

        # Base buttons
        self.plot_without_asr_dipdip_btn = pnw.Button(name="Compute phonons with/wo ASR and DIPDIP",
                                                      button_type='primary')

        #if self.has_remote_serve:
        #    self.param.nqsmall.bounds = (1, 50)
        #    self.param.ndivsm.bounds = (-30, 30)

    def kwargs_for_anaget_phbst_and_phdos_files(self, **extra_kwargs) -> dict:
        """
        Return the parameters required to invoke anaget_phbst_and_phdos_files
        Additional kwargs can be specified in extra_kwargs if needed.
        """
        d = dict(nqsmall=self.nqsmall, qppa=None, ndivsm=self.ndivsm,
                 line_density=None, asr=self.asr, chneut=self.chneut, dipdip=self.dipdip,
                 dipquad=self.dipquad, quadquad=self.quadquad,
                 dos_method=self.dos_method, lo_to_splitting=self.lo_to_splitting,
                 verbose=self.verbose, mpi_procs=self.mpi_procs)

        if extra_kwargs: d.update(extra_kwargs)

        return d


class DdbFilePanel(PanelWithStructure, PanelWithAnaddbParams):
    """
    A panel to analyze a |DdbFile|.
    Provides widgets to invoke anaddb and visualize the results.
    """

    vsound_num_points = param.Integer(20, bounds=(1, None), label="Number of q-points")

    vsound_qpt_norm = param.Number(0.1, bounds=(0, None), label="Norm of the largest q-point")

    def __init__(self, ddb: DdbFile, **params):
        PanelWithStructure.__init__(self, structure=ddb.structure, **params)
        PanelWithAnaddbParams.__init__(self)
        self.ddb = ddb

        # Add buttons
        self.get_epsinf_btn = pnw.Button(name="Compute", button_type='primary')
        self.plot_phbands_btn = pnw.Button(name="Plot Bands and DOS", button_type='primary')
        self.plot_eps0w_btn = pnw.Button(name="Plot eps0(omega)", button_type='primary')

        self.plot_vsound_btn = pnw.Button(name="Calculate speed of sound", button_type='primary')
        self.plot_ifc_btn = pnw.Button(name="Compute IFC(R)", button_type='primary')
        self.plot_phbands_quad_btn = pnw.Button(name="Plot PHbands with/without quadrupoles", button_type='primary')
        self.plot_dos_vs_qmesh_btn = pnw.Button(name="Plot PHDos vs Qmesh", button_type='primary')
        self.compute_elastic_btn = pnw.Button(name="Compute Elastic", button_type='primary')

        self.stacked_pjdos = pnw.Checkbox(name="Stacked PJDOS", value=True)
        self.with_qpath = pnw.Checkbox(name="Show q-path with plotly", value=True)

    def get_becs_view(self) -> pn.Row:
        return pn.Row(
            self.pws_col(["## Born effective charges options",
                          "asr", "chneut", "dipdip", "eps0_gamma_ev", "get_epsinf_btn",
                          ]),
            self.get_epsinf
        )

    @depends_on_btn_click('get_epsinf_btn')
    def get_epsinf(self) -> pn.Column:
        """
        Compute eps_infinity and Born effective charges from DDB.
        """
        epsinf, becs = self.ddb.anaget_epsinf_and_becs(chneut=self.chneut,
                                                       mpi_procs=self.mpi_procs, verbose=self.verbose)

        gen, inp = self.ddb.anaget_dielectric_tensor_generator(asr=self.asr, chneut=self.chneut, dipdip=self.dipdip,
                                                               mpi_procs=self.mpi_procs, verbose=self.verbose,
                                                               return_input=True)

        # Fill column
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append
        eps0 = gen.tensor_at_frequency(w=0, gamma_ev=self.eps0_gamma_ev)

        df_kwargs = {}
        def m(s):
            return pn.pane.Markdown(s)
            #return pn.Row(pn.pane.LaTeX(s, style={'font-size': '18pt'}), sizing_mode="stretch_width")

        #ca(m(f"$$\epsilon_0$$ in Cart. coords. Computed with gamma_eV: {self.eps0_gamma_ev:.2f}"))
        ca(m(r"## Cartesian coordinates of $$\epsilon_0$$ tensor"))
        ca(dfc(eps0.get_dataframe(cmode="real"), **df_kwargs))
        ca(m(r"## Cartesian coordinates of $$\epsilon_\infty$$ tensor"))
        ca(dfc(epsinf.get_dataframe(), **df_kwargs))
        ca("## Born effective charges in Cartesian coordinates:")
        ca(dfc(becs.get_voigt_dataframe(), **df_kwargs))
        ca("## Anaddb input file")
        ca(self.html_with_clipboard_btn(inp))

        #pn.state.notifications.success('This is a success notification.', duration=0)
        #pn.state.notifications.send('Fire!!!', background='red', icon='<i class="fas fa-burn"></i>')
        #return pn.io.notifications.NotificationArea.demo()

        return col

    def get_eps0_view(self) -> pn.Row:
        return pn.Row(
                self.pws_col(["## epsilon_0",
                              "asr", "chneut", "dipdip", "eps0_gamma_ev", "eps0_wrange", "plot_eps0w_btn",
                             ]),
                self.plot_eps0w
            )

    @depends_on_btn_click('plot_eps0w_btn')
    def plot_eps0w(self) -> pn.Column:
        """
        Compute eps0(omega) from DDB and plot the results.
        """
        gen, inp = self.ddb.anaget_dielectric_tensor_generator(asr=self.asr, chneut=self.chneut, dipdip=self.dipdip,
                                                               mpi_procs=self.mpi_procs, verbose=self.verbose,
                                                               return_input=True)
        ws = self.eps0_wrange.value
        w_max = ws[1]
        if w_max == 1.0: w_max = None # Will compute w_max in plot routine from ph freqs.

        def p(component, reim):
            fig = gen.plotly(w_min=ws[0], w_max=w_max, gamma_ev=self.eps0_gamma_ev, num=500, component=component,
                              reim=reim, units=self.units, show=False)
            return ply(fig, with_help=False)

        col = pn.Column(sizing_mode='stretch_width'); ca = col.append

        # Add figures
        ca("## epsilon(w):")
        ca(p("diag", "re"))
        ca(p("diag", "im"))
        ca(p("offdiag", "re"))
        ca(p("offdiag", "im"))

        #gspec[2, :] = gen.get_oscillator_dataframe(reim="all", tol=1e-6)
        # TODO: FIX
        # TypeError: Object of type complex is not JSON serializable
        #dfc(gen.get_oscillator_dataframe(reim="all", tol=1e-6))
        ca("## Oscillator matrix elements:")
        ca(gen.get_oscillator_dataframe(reim="all", tol=1e-6))
        # Add HTML pane with input.
        ca("## Anaddb input file")
        ca(self.html_with_clipboard_btn(inp))

        return col

    def get_phbands_view(self) -> pn.Row:
        return pn.Row(
                self.pws_col(["## PH-bands options",
                              "nqsmall", "ndivsm", "asr", "chneut",
                              "dipdip", "dipquad", "quadquad",
                              "lo_to_splitting", "dos_method",
                              "stacked_pjdos", "with_qpath", "temp_range",
                              pn.layout.Divider(),
                              "plot_phbands_btn",
                             ]),
                self.on_plot_phbands_and_phdos
            )

    @depends_on_btn_click('plot_phbands_btn')
    @add_lr_docstring
    def on_plot_phbands_and_phdos(self) -> pn.Column:
        """
        This Tab provides widgets to compute:

        - Phonon dispersion along a high-symmetry **q**-path
        - Total phonon DOS and atom-projected phonon DOSes
        - Thermodynamic properties in the Harmonic approximation

        Important

            Note that **anaddb** uses the **q**-mesh found in the DDB file to build the
            interatomic force constants (IFCs) in real space and then compute phonon quantities
            at arbitrary **q**-points by Fourier interpolation.

        """
        # Computing phbands
        kwargs = self.kwargs_for_anaget_phbst_and_phdos_files(return_input=True)

        with self.ddb.anaget_phbst_and_phdos_files(**kwargs) as g:
            phbst_file, phdos_file = g
            phbands, phdos = phbst_file.phbands, phdos_file.phdos

            # Fill column
            col = pn.Column(sizing_mode='stretch_width'); ca = col.append

            ca("## Phonon band structure and DOS:")
            ca(ply(phbands.plotly_with_phdos(phdos, units=self.units, show=False)))

            if self.with_qpath.value:
                ca("## Brillouin zone and q-path:")
                qpath_pane = ply(phbands.qpoints.plotly(show=False), with_divider=False)
                df_qpts = phbands.qpoints.get_highsym_datataframe()
                ca(pn.Row(qpath_pane, df_qpts))
                ca(pn.layout.Divider())

            ca("## Type-projected phonon DOS:")
            ca(ply(phdos_file.plotly_pjdos_type(units=self.units, stacked=self.stacked_pjdos.value, show=False)))

            ca("## Thermodynamic properties in the harmonic approximation:")
            temps = self.temp_range.value
            ca(ply(phdos.plotly_harmonic_thermo(tstart=temps[0], tstop=temps[1], num=50, show=False)))
            #ca(mpl(phdos_file.msqd_dos.plot(units=self.units, **self.mpl_kwargs)))
            #msqd_dos.plot_tensor(**self.mpl_kwargs)

            # Add Anaddb input file
            ca("## Anaddb input file")
            ca(self.html_with_clipboard_btn(g.input))

            return col

    def get_vsound_view(self) -> pn.Row:
        return pn.Row(
                self.pws_col(["## Speed of sound options",
                              "vsound_num_points", "vsound_qpt_norm",
                              "asr", "chneut", "dipdip", "dipquad", "quadquad",
                              "plot_vsound_btn",
                            ]),
                self.plot_vsound
            )

    @depends_on_btn_click('plot_vsound_btn')
    @add_lr_docstring
    def plot_vsound(self) -> pn.Column:
        """
        This Tab provides widgets to compute the speed of sound by fitting
        the **first three** phonon branches along selected **q**-directions by
        linear least-squares fit.
        """
        col = pn.Column(sizing_mode="stretch_width"); ca = col.append

        from abipy.dfpt.vsound import SoundVelocity
        sv, inp = SoundVelocity.from_ddb(self.ddb.filepath, num_points=self.vsound_num_points, qpt_norm=self.vsound_qpt_norm,
                                         ignore_neg_freqs=True, asr=self.asr, chneut=self.chneut,
                                         dipdip=self.dipdip, dipquad=self.dipquad, quadquad=self.quadquad,
                                         verbose=self.verbose, mpi_procs=self.mpi_procs, return_input=True)

        ca("## Linear least-squares fit")
        ca(ply(sv.plotly(show=False)))
        ca("## Speed of sound computed along different q-directions in reduced coords")
        ca(dfc(sv.get_dataframe()))
        ca("## Anaddb input file")
        ca(self.html_with_clipboard_btn(inp))

        return col

    def get_asr_dipdip_view(self) -> pn.Row:
        return pn.Row(
                    self.pws_col(["## ASR & DIPDIP options",
                                  "nqsmall", "ndivsm", "dos_method", "plot_without_asr_dipdip_btn",
                                 ]),
                    self.plot_without_asr_dipdip
                )

    @depends_on_btn_click('plot_without_asr_dipdip_btn')
    def plot_without_asr_dipdip(self) -> pn.Column:
        """
        This Tab provides widgets to compare phonon bands and DOSes computed with/without
        enforcing the acoustic sum rule and the treatment of the dipole-dipole interaction in the dynamical matrix.
        Requires DDB file with eps_inf, BECS.
        """
        asr_plotter = self.ddb.anacompare_asr(asr_list=(0, 2), chneut_list=(1, ), dipdip=1, dipquad=1, quadquad=1,
                                              lo_to_splitting=self.lo_to_splitting,
                                              nqsmall=self.nqsmall, ndivsm=self.ndivsm,
                                              dos_method=self.dos_method, ngqpt=None,
                                              verbose=self.verbose, mpi_procs=self.mpi_procs)

        dipdip_plotter = self.ddb.anacompare_dipdip(chneut_list=(1,), asr=2, lo_to_splitting=self.lo_to_splitting,
                                                    nqsmall=self.nqsmall, ndivsm=self.ndivsm,
                                                    dos_method=self.dos_method, ngqpt=None,
                                                    verbose=self.verbose, mpi_procs=self.mpi_procs)

        # Fill column
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append
        ca("## Phonon bands and DOS with/wo acoustic sum rule:")
        ca(ply(asr_plotter.combiplotly(show=False)))
        ca("## Phonon bands and DOS with/without the treatment of the dipole-dipole interaction:")
        ca(ply(dipdip_plotter.combiplotly(show=False)))

        return col

    def get_dos_vs_qmesh_view(self) -> pn.Row:
        return pn.Row(
                self.pws_col(["## DOS vs q-mesh options",
                              "nqsmall_list", "dos_method",
                              "asr", "chneut", "dipdip", "dipquad", "quadquad", "temp_range",
                              "plot_dos_vs_qmesh_btn",
                             ]),
                self.plot_dos_vs_qmesh
            )

    @depends_on_btn_click('plot_dos_vs_qmesh_btn')
    @add_lr_docstring
    def plot_dos_vs_qmesh(self) -> pn.Column:
        """
        This Tab provides widgets to compare phonon DOSes and thermodynamic properties
        obtained using different q-meshes specifined via `nqsmall_list`.
        """
        r = self.ddb.anacompare_phdos(self.nqsmall_list.value, asr=self.asr, chneut=self.chneut,
                                      dipdip=self.dipdip, dipquad=self.dipquad, quadquad=self.quadquad,
                                      dos_method=self.dos_method, ngqpt=None,
                                      verbose=self.verbose, num_cpus=1, stream=sys.stdout)

        #r.phdoses: List of |PhononDos| objects

        # Fill column
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append
        ca("## Phonon DOSes obtained with different q-meshes:")
        ca(ply(r.plotter.combiplotly(show=False)))

        ca("## Convergence of termodynamic properties.")
        temps = self.temp_range.value
        #ca(mpl(r.plotter.plot_harmonic_thermo(tstart=temps[0], tstop=temps[1], num=50,
        #                                      units=self.units, **self.mpl_kwargs)))
        ca(ply(r.plotter.plotly_harmonic_thermo(tstart=temps[0], tstop=temps[1], num=50,
                                                units=self.units, show=False)))

        return col

    def get_quadrupoles_view(self) -> pn.Row:
        return pn.Row(
            self.pws_col(["## Quadrupoles options",
                          "asr", "chneut", "dipdip", "lo_to_splitting", "ndivsm", "dos_method",
                          "plot_phbands_quad_btn",
                          ]),
            self.plot_phbands_quad
        )

    @depends_on_btn_click('plot_phbands_quad_btn')
    def plot_phbands_quad(self) -> pn.Column:
        """
        This Tab provides widgets to compare phonon bands and DOSes computed with/without the inclusion
        of the dipole-quadrupole and quadrupole-quadrupole terms in the dynamical matrix.
        Requires DDB file with eps_inf, BECS and dynamical quadrupoles.
        """
        plotter = self.ddb.anacompare_quad(asr=self.asr, chneut=self.chneut, dipdip=self.dipdip,
                                           lo_to_splitting=self.lo_to_splitting,
                                           nqsmall=0, ndivsm=self.ndivsm, dos_method=self.dos_method, ngqpt=None,
                                           verbose=self.verbose, mpi_procs=self.mpi_procs)

        # Fill column
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append
        ca("## Phonon Bands obtained with different q-meshes:")
        ca(ply(plotter.combiplotly(show=False)))

        return col

    def get_ifcs_view(self) -> pn.Row:
        return pn.Row(
                self.pws_col(["## IFCs options",
                               "asr", "dipdip", "chneut", "ifc_yscale", "plot_ifc_btn",
                             ]),
                self.on_plot_ifc
            )

    @depends_on_btn_click('plot_ifc_btn')
    def on_plot_ifc(self):
        """
        Plot the Interatomic Force Constants in real space.
        """
        ifc, inp = self.ddb.anaget_ifc(asr=self.asr, chneut=self.chneut, dipdip=self.dipdip, return_input=True)

        kwds = self.mpl_kwargs.copy()
        kwds["yscale"] = self.ifc_yscale
        #print(kwds)

        # Fill column
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append
        ca(mpl(ifc.plot_longitudinal_ifc(title="Longitudinal IFCs", **kwds)))
        ca(mpl(ifc.plot_longitudinal_ifc_short_range(title="Longitudinal IFCs short range", **kwds)))
        ca(mpl(ifc.plot_longitudinal_ifc_ewald(title="Longitudinal IFCs Ewald", **kwds)))
        ca("## Anaddb input file")
        ca(self.html_with_clipboard_btn(inp))

        return col

    def get_elastic_view(self) -> pn.Row:
        return pn.Row(
            self.pws_col(["## Elastic options",
                          "asr", "chneut", "compute_elastic_btn",
                          ]),
            self.on_compute_elastic_btn
        )

    @depends_on_btn_click('compute_elastic_btn')
    def on_compute_elastic_btn(self) -> pn.Column:
        """
        Call anaddb to compute elastic and piezoelectric tensors. Require DDB with strain terms.

        By default, this method sets the anaddb input variables automatically
        by looking at the 2nd-order derivatives available in the DDB file.
        This behaviour can be changed by setting explicitly the value of:
        `relaxed_ion` and `piezo`.
        """

        edata, inp = self.ddb.anaget_elastic(relaxed_ion="automatic", piezo="automatic",
                                             dde=False, stress_correction=False, asr=self.asr, chneut=self.chneut,
                                             mpi_procs=1, verbose=self.verbose, return_input=True)

        col = pn.Column(sizing_mode='stretch_width'); ca = col.append

        #print(edata)
        #edata.elastic_relaxed
        #edata.elastic_relaxed.compliance_tensor
        #edata.get_elastic_properties_dataframe(properties_as_index=True)
        #edata.get_elastic_voigt_dataframe(tol=1e-5)
        #edata.piezo_relaxed

        # This will open a new tab in the browser with the ELATE webpage
        #html = edata.elastic_relaxed.get_elate_html(sysname="FooBar")
        #return pn.Template(html).show()

        ca("## Anaddb input file")
        ca(self.html_with_clipboard_btn(inp))

        return col

    def get_panel(self, as_dict=False, **kwargs):
        """
        Return tabs with widgets to interact with the DDB file.
        """
        ddb = self.ddb
        d = {}
        d["Summary"] = self.get_summary_view_for_abiobj(self.ddb)

        # Note how we build tabs according to the content of the DDB.
        if ddb.has_at_least_one_atomic_perturbation():
            d["PH-bands"] = self.get_phbands_view()

        if ddb.has_bec_terms(select="at_least_one"):
            d["BECs"] = self.get_becs_view()

        if ddb.has_epsinf_terms(select="at_least_one_diagoterm"):
            d["eps0"] = self.get_eps0_view()

        if ddb.has_at_least_one_atomic_perturbation():
            d["Speed of sound"] = self.get_vsound_view()
            d["DOS vs q-mesh"] = self.get_dos_vs_qmesh_view()

            if ddb.has_lo_to_data():
                d["ASR & DIPDIP"] = self.get_asr_dipdip_view()
            if ddb.has_quadrupole_terms():
                d["Quadrupoles"] = self.get_quadrupoles_view()

            d["IFCs"] = self.get_ifcs_view()

        if self.ddb.has_strain_terms():
            d["Elastic"] = self.get_elastic_view()

        d["Structure"] = self.get_structure_view()
        d["Global"] = pn.Row(
            self.pws_col(["## Global options", "units", "mpi_procs", "verbose"]),
            self.get_software_stack()
        )

        if as_dict: return d
        return self.get_template_from_tabs(d, template=kwargs.get("template", None))


class PanelWithFileInput(AbipyParameterized):

    info_str = """
Post-process the data stored in one of the ABINIT output files (*GSR.nc*, *DDB*, *FATBANDS.nc*, etc)
"""

    def __init__(self, use_structure=False, **params):

        super().__init__(**params)

        self.use_structure = use_structure
        help_md = pn.pane.Markdown(f"""
## Description

{self.info_str}

Use the **Choose File** to upload one of the files supported by this app.
Drop one of of the files supported by AbiPy onto the FileInput area or
click the **Choose File** button to upload
Keep in mind that the **file extension matters**!
Also, avoid uploading big files (size > XXX).
""")

        self.main_area = pn.Column(help_md, sizing_mode="stretch_width")
        self.abifile = None

        self.file_input = pnw.FileInput(height=60, css_classes=["pnx-file-upload-area"])
        self.file_input.param.watch(self.on_file_input, "value")

        self.mpid_input = pnw.TextInput(name='mp-id', placeholder='Enter e.g. mp-149 for Silicon and press ⏎')
        self.mpid_input.param.watch(self.on_mpid_input, "value")
        self.mpid_err_wdg = pn.pane.Markdown("")
        #self.mp_progress = pn.indicators.Progress(name='Fetching data from the MP website', bar_color="warning",
        #                                          active=False, width=200, height=10, align="center")

    def on_file_input(self, event) -> None:

        with Loading(self.main_area):
            new_abifile = self.get_abifile_from_file_input(self.file_input, use_structure=self.use_structure)

            if self.abifile is not None:
                self.abifile.remove()

            self.abifile = new_abifile
            self.main_area.objects = [self.abifile.get_panel()]

    def on_mpid_input(self, event) -> None:

        with Loading(self.mpid_input, err_wdg=self.mpid_err_wdg):
            self.abifile = Structure.from_mpid(self.mpid_input.value)

            self.main_area.objects = [self.abifile.get_panel()]

    def get_panel(self):

        if self.use_structure:
            title = "Structure Analyzer"
            msg = "## Upload (or drag & drop) **any file** with a structure (*.nc*, *.abi*, *.cif*, *.xsf*, POSCAR):"
        else:
            title = "Output File Analyzer"
            msg = "## Upload (or drag & drop) **any file** supported by AbiPy. See list below:"

        col = pn.Column(
            msg,
            self.get_fileinput_section(self.file_input),
            self.wdg_exts_with_get_panel(),
            sizing_mode="stretch_width")

        if self.use_structure:
            col.extend([
                "## or get the structure from the [Materials Project](https://materialsproject.org/) database:",
                pn.Row(self.mpid_input, pn.Column(self.mpid_err_wdg), sizing_mode="stretch_width"),
            ])

        main = pn.Column(col, self.main_area, sizing_mode="stretch_width")

        cls, kwds = self.get_abinit_template_cls_kwds()

        return cls(main=main, title=title, **kwds)


class PanelWithStructureInput(PanelWithFileInput):

    info_str = """
This app allows users to upload a file with structure info and operate on it.
"""

    def __init__(self, **params):
        super().__init__(use_structure=True, **params)


class DdbPanelWithFileInput(AbipyParameterized):

    info_str = """
This app allows users to upload a DDB file with the dynamical matrix, compute
phonon-related properties such as band structures, DOS, infrared absorption and visualize
the results.
"""

    def __init__(self, **params):

        super().__init__(**params)

        help_md = pn.pane.Markdown(f"""
## Description

{self.info_str}
""")

        self.main_area = pn.Column(help_md,
                                   self.get_alert_data_transfer(),
                                   sizing_mode="stretch_width")
        self.abifile = None

        self.file_input = pnw.FileInput(height=60, css_classes=["pnx-file-upload-area"])
        self.file_input.param.watch(self.on_file_input, "value")

        self.mpid_input = pnw.TextInput(name='mp-id', placeholder='Enter e.g. mp-149 for Silicon and press ⏎')
        self.mpid_input.param.watch(self.on_mpid_input, "value")
        self.mpid_err_wdg = pn.pane.Markdown("")
        #self.mp_progress = pn.indicators.Progress(name='Fetching data from the MP website', bar_color="warning",
        #                                          active=False, width=200, height=10, align="center")

    def on_file_input(self, event) -> None:

        with Loading(self.main_area):
            self.mpid_err_wdg.object = ""
            new_abifile = self.get_abifile_from_file_input(self.file_input)

            if self.abifile is not None:
                self.abifile.remove()

            self.abifile = new_abifile
            self.main_area.objects = [self.abifile.get_panel()]

    def on_mpid_input(self, event):

        from abipy.dfpt.ddb import DdbFile
        with Loading(self.mpid_input, err_wdg=self.mpid_err_wdg):
            self.abifile = DdbFile.from_mpid(self.mpid_input.value)
            self.main_area.objects = [self.abifile.get_panel()]

    def get_panel(self):

        col = pn.Column(
            "## Upload (or drag & drop) a DDB file:",
            self.get_fileinput_section(self.file_input),
            "## or get the DDB from the [Materials Project](https://materialsproject.org/) database (*if available*):",
            pn.Row(self.mpid_input, pn.Column(self.mpid_err_wdg), sizing_mode="stretch_width"),
            sizing_mode="stretch_width")

        main = pn.Column(col, self.main_area, sizing_mode="stretch_width")
        cls, kwds = self.get_abinit_template_cls_kwds()

        return cls(main=main, title="DDB Analyzer", **kwds)


class CompareDdbWithMP(AbipyParameterized):

    info_str = """
This app alllows users to upload a DDB file and compare it with the one available on the MP.
"""

    def __init__(self, **params):

        super().__init__(**params)

        help_md = pn.pane.Markdown(f"""
## Description

{self.info_str}
""")

        self.main_area = pn.Column(help_md,
                                   self.get_alert_data_transfer(),
                                   sizing_mode="stretch_width")

        self.file_input = pnw.FileInput(height=60, css_classes=["pnx-file-upload-area"])
        self.file_input.param.watch(self.on_file_input, "value")

        self.mp_progress = pn.indicators.Progress(name='Fetching DDB from the MP website', bar_color="warning",
                                                  active=False, width=100, height=10, align="center")
        self.mp_err_wdg = pn.pane.Markdown("")

    def on_file_input(self, event) -> None:
        abinit_ddb = self.get_abifile_from_file_input(self.file_input)
        from abipy.dfpt.ddb import DdbFile, DdbRobot

        # Match Abinit structure with MP.
        mp = abinit_ddb.structure.mp_match()

        with ActiveBar(self.mp_progress, err_wdg=self.mp_err_wdg), Loading(self.main_area):
            mpid_list = [mp_id for mp_id in mp.ids if mp_id != "this"]
            ddb_robot = DdbRobot.from_mpid_list(mpid_list)
            ddb_robot.add_file("Yours DDB", abinit_ddb)

            self.main_area.objects = [DdbRobotPanel(ddb_robot).get_panel()]

    def get_panel(self):
        col = pn.Column(
            "## Upload (or drag & drop) a DDB file:",
            self.get_fileinput_section(self.file_input),
            pn.Row("## Fetching data from the MP website: ", self.mp_progress, self.mp_err_wdg,
                   sizing_mode="stretch_width"),
            sizing_mode="stretch_width")

        main = pn.Column(col, self.main_area, sizing_mode="stretch_width")
        cls, kwds = self.get_abinit_template_cls_kwds()

        return cls(main=main, title="Compare with MP DDB", **kwds)


class DdbRobotPanel(BaseRobotPanel, PanelWithAnaddbParams):
    """
    A panel to analyze multiple DdbFiles via the low-level API provided by DdbRobot.
    Provides widgets to invoke anaddb and visualize the results.
    """
    def __init__(self, robot: DdbRobot, **params):
        BaseRobotPanel.__init__(self, robot=robot, **params)
        PanelWithAnaddbParams.__init__(self)

        # Buttons
        self.plot_combiplot_btn = pnw.Button(name="Compute", button_type='primary')
        self.combiplot_check_btn = pnw.CheckButtonGroup(name='Check Button Group',
                                                        value=['combiplot'], options=['combiplot', 'gridplot'])

    def kwargs_for_anaget_phbst_and_phdos_files(self, **extra_kwargs) -> dict:
        """Extend method of base class to handle lo_to_splitting"""
        kwargs = super().kwargs_for_anaget_phbst_and_phdos_files(**extra_kwargs)

        if kwargs["lo_to_splitting"] == "automatic":
            if any(not ddb.has_lo_to_data() for ddb in self.robot.abifiles):
                kwargs["lo_to_splitting"] = False
                if self.verbose:
                    print("Setting lo_to_splitting to False since at least one DDB file does not have LO-TO data.")

        return kwargs

    @depends_on_btn_click('plot_combiplot_btn')
    def plot_combiplot(self, **kwargs) -> pn.Column:
        """Plot phonon band structures."""
        kwargs = self.kwargs_for_anaget_phbst_and_phdos_files()

        #TODO: Recheck lo-to automatic.
        r = self.robot.anaget_phonon_plotters(**kwargs)
        #r = self.robot.anaget_phonon_plotters()

        # Fill column
        col = pn.Column(sizing_mode='stretch_both'); ca = col.append

        if "combiplot" in self.combiplot_check_btn.value:
            ca("## Combiplot:")
            ca(ply(r.phbands_plotter.combiplotly(units=self.units, show=False)))

        if "gridplot" in self.combiplot_check_btn.value:
            ca("## Gridplot:")
            # FIXME implement with_dos = True
            ca(ply(r.phbands_plotter.gridplotly(units=self.units, with_dos=False, show=False)))

        #if "temp_range" in self.combiplot_check_btn.value:
        #temps = self.temp_range.value
        #ca("## Thermodynamic properties in the harmonic approximation:")
        ##ca(phdos.plot_harmonic_thermo(tstart=temps[0], tstop=temps[1], num=50, **self.mpl_kwargs))
        #ca(ply(phdos.plotly_harmonic_thermo(tstart=temps[0], tstop=temps[1], num=50, show=False)))

        return col

    #@param.depends('get_epsinf_btn.clicks')
    #def get_epsinf(self):
    #    """Compute eps_infinity and Born effective charges from DDB."""
    #    if self.get_epsinf_btn.clicks == 0: return

    #    with ButtonContext(self.get_epsinf_btn):
    #        epsinf, becs = self.ddb.anaget_epsinf_and_becs(chneut=self.chneut,
    #                                                       mpi_procs=self.mpi_procs, verbose=self.verbose)

    #        gen, inp = self.ddb.anaget_dielectric_tensor_generator(asr=self.asr, chneut=self.chneut, dipdip=self.dipdip,
    #                                                               mpi_procs=self.mpi_procs, verbose=self.verbose,
    #                                                               return_input=True)

    #        # Fill column
    #        col = pn.Column(sizing_mode='stretch_width'); ca = col.append
    #        df_kwargs = dict(auto_edit=False, autosize_mode="fit_viewport")
    #        #l = pn.pane.LaTeX

    #        eps0 = gen.tensor_at_frequency(w=0, gamma_ev=self.eps0_gamma_ev)
    #        ca(r"## $\epsilon^0$ in Cart. coords (computed with Gamma_eV):")
    #        ca(dfc(eps0.get_dataframe(cmode="real"), **df_kwargs))
    #        ca(r"## $\epsilon^\infty$ in Cart. coords:")
    #        ca(dfc(epsinf.get_dataframe(), **df_kwargs))
    #        ca("## Born effective charges in Cart. coords:")
    #        ca(dfc(becs.get_voigt_dataframe(), **df_kwargs))
    #        ca("## Anaddb input file")
    #        ca(self.html_with_clipboard_btn(inp))

    #        return col

    #@param.depends('plot_eps0w_btn.clicks')
    #def plot_eps0w(self):
    #    """Compute eps0(omega) from DDB and plot the results."""
    #    if self.plot_eps0w_btn.clicks == 0: return

    #    with ButtonContext(self.plot_eps0w_btn):
    #        gen, inp = self.ddb.anaget_dielectric_tensor_generator(asr=self.asr, chneut=self.chneut, dipdip=self.dipdip,
    #                                                               mpi_procs=self.mpi_procs, verbose=self.verbose,
    #                                                               return_input=True)
    #        ws = self.eps0_wrange.value
    #        w_max = ws[1]
    #        if w_max == 1.0: w_max = None # Will compute w_max in plot routine from ph freqs.

    #        def p(component, reim):
    #            return gen.plot(w_min=ws[0], w_max=w_max, gamma_ev=self.eps0_gamma_ev, num=500, component=component,
    #                            reim=reim, units=self.units, **self.mpl_kwargs)

    #        # Build grid
    #        gspec = pn.GridSpec(sizing_mode='scale_width')
    #        gspec[0, 0] = p("diag", "re")
    #        gspec[0, 1] = p("diag", "im")
    #        gspec[1, 0] = p("offdiag", "re")
    #        gspec[1, 1] = p("offdiag", "im")
    #        gspec[2, :] = gen.get_oscillator_dataframe(reim="all", tol=1e-6)
    #        # Add HTML pane with input.
    #        gspec[3, 0] = pn.pane.HTML(inp._repr_html_())
    #        ca(self.html_with_clipboard_btn(inp))

    #        return gspec

    #@param.depends('plot_phbands_btn.clicks')
    #def on_plot_phbands_and_phdos(self, event=None):
    #    """Compute phonon bands and DOSes from DDB and plot the results."""
    #    if self.plot_phbands_btn.clicks == 0: return

    #    with ButtonContext(self.plot_phbands_btn):
    #        # Computing phbands
    #        with self.ddb.anaget_phbst_and_phdos_files(
    #                nqsmall=self.nqsmall, qppa=None, ndivsm=self.ndivsm,
    #                line_density=None, asr=self.asr, chneut=self.chneut, dipdip=self.dipdip,
    #                dos_method=self.dos_method, lo_to_splitting=self.lo_to_splitting,
    #                verbose=self.verbose, mpi_procs=self.mpi_procs, return_input=True) as g:

    #            phbst_file, phdos_file = g
    #            phbands, phdos = phbst_file.phbands, phdos_file.phdos

    #        # Fill column
    #        col = pn.Column(sizing_mode='stretch_width'); ca = col.append

    #        ca("## Phonon band structure and DOS:")
    #        ca(ply(phbands.plotly_with_phdos(phdos, units=self.units, show=False)))
    #        #ca(mpl(phbands.plot_with_phdos(phdos, units=self.units, **self.mpl_kwargs)))
    #        #ca(mpl(phdos_file.plot_pjdos_type(units=self.units, exchange_xy=True, **self.mpl_kwargs)))
    #        #ca(mpl(phdos_file.msqd_dos.plot(units=self.units, **self.mpl_kwargs)))
    #        temps = self.temp_range.value
    #        ca("## Thermodynamic properties in the harmonic approximation:")
    #        #ca(phdos.plot_harmonic_thermo(tstart=temps[0], tstop=temps[1], num=50, **self.mpl_kwargs))
    #        ca(ply(phdos.plotly_harmonic_thermo(tstart=temps[0], tstop=temps[1], num=50, show=False)))
    #        #msqd_dos.plot_tensor(**self.mpl_kwargs)
    #        #self.plot_phbands_btn.button_type = "primary"

    #        # Add HTML pane with input
    #        ca("## Anaddb input file")
    #        ca(self.html_with_clipboard_btn(inp))

    #        return col

    #@param.depends('plot_vsound_btn.clicks')
    #def plot_vsound(self):
    #    """
    #    Compute the speed of sound by fitting phonon frequencies
    #    along selected directions by linear least-squares fit.
    #    """
    #    if self.plot_vsound_btn.clicks == 0: return

    #    with ButtonContext(self.plot_vsound_btn):
    #        from abipy.dfpt.vsound import SoundVelocity
    #        sv = SoundVelocity.from_ddb(self.ddb.filepath, num_points=20, qpt_norm=0.1,
    #                                    ignore_neg_freqs=True, asr=self.asr, chneut=self.chneut, dipdip=self.dipdip,
    #                                    verbose=self.verbose, mpi_procs=self.mpi_procs)

    #        # Insert results in grid.
    #        gspec = pn.GridSpec(sizing_mode='scale_width')
    #        gspec[0, :1] = sv.get_dataframe()
    #        gspec[1, :1] = sv.plot(**self.mpl_kwargs)

    #        return gspec

    # THIS OK but I don't think it's very useful
    @depends_on_btn_click('plot_without_asr_dipdip_btn')
    def plot_without_asr_dipdip(self) -> pn.Column:
        """
        Compare phonon bands and DOSes computed with/without the acoustic sum rule
        and the treatment of the dipole-dipole interaction in the dynamical matrix.
        Requires DDB file with eps_inf, BECS.
        """
        asr_plotter = PhononBandsPlotter()
        dipdip_plotter = PhononBandsPlotter()

        for label, ddb in self.robot.items():
            asr_p = ddb.anacompare_asr(asr_list=(0, 2), chneut_list=(1, ), dipdip=1,
                                       lo_to_splitting=self.lo_to_splitting,
                                       nqsmall=self.nqsmall, ndivsm=self.ndivsm,
                                       dos_method=self.dos_method, ngqpt=None,
                                       verbose=self.verbose, mpi_procs=self.mpi_procs,
                                       pre_label=label)

            asr_plotter.append_plotter(asr_p)

            dipdip_p = ddb.anacompare_dipdip(chneut_list=(1,), asr=2, lo_to_splitting=self.lo_to_splitting,
                                             nqsmall=self.nqsmall, ndivsm=self.ndivsm,
                                             dos_method=self.dos_method, ngqpt=None,
                                             verbose=self.verbose, mpi_procs=self.mpi_procs,
                                             pre_label=label)

            dipdip_plotter.append_plotter(dipdip_p)

        # Fill column
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append

        ca("## Phonon bands and DOS with/wo the acoustic sum rule:")
        ca(ply(asr_plotter.combiplotly(show=False)))
        ca("## Phonon bands and DOS with/without the treatment of the dipole-dipole interaction:")
        ca(ply(dipdip_plotter.combiplotly(show=False)))

        return col

    def get_panel(self, as_dict=False, **kwargs):
        """Return tabs with widgets to interact with the DDB file."""
        robot = self.robot

        d = {}
        d["Summary"] = self.get_summary_view_for_abiobj(robot)
        d["Params"] = self.get_compare_params_widgets()

        d["Plot"] = pn.Row(
            self.pws_col(["# PH-bands options",
                           "nqsmall", "ndivsm", "asr", "chneut", "dipdip",
                           "lo_to_splitting", "dos_method", "temp_range",
                           "combiplot_check_btn", "plot_combiplot_btn",
                         ]),
            self.plot_combiplot
        )
        #app(("PH-bands", pn.Row(
        #    pn.Column("# PH-bands options",
        #              *self.pws("nqsmall", "ndivsm", "asr", "chneut", "dipdip",
        #                        "lo_to_splitting", "dos_method", "temp_range", "plot_phbands_btn",
        #                        self.helpc("on_plot_phbands_and_phdos")),
        #              ),
        #    self.on_plot_phbands_and_phdos)
        #))
        #app(("BECs", pn.Row(
        #    pn.Column("# Born effective charges options",
        #              *self.pws("asr", "chneut", "dipdip", "eps0_gamma_ev", "get_epsinf_btn",
        #                        self.helpc("get_epsinf")),
        #             ),
        #    self.get_epsinf)
        #))
        #app(("eps0", pn.Row(
        #    pn.Column("# epsilon_0",
        #              *self.pws("asr", "chneut", "dipdip", "eps0_gamma_ev", "eps0_wrange", "plot_eps0w_btn",
        #                        self.helpc("plot_eps0w")),
        #              ),
        #    self.plot_eps0w)
        #))
        #app(("Speed of sound", pn.Row(
        #    pn.Column("# Speed of sound options",
        #              *self.pws("asr", "chneut", "dipdip", "plot_vsound_btn",
        #                        self.helpc("plot_vsound")),
        #              ),
        #    self.plot_vsound)
        #))
        #d["ASR & DIPDIP"] = pn.Row(
        #    self.pws_col(["## ASR & DIPDIP options", "nqsmall", "ndivsm", "dos_method", "plot_without_asr_dipdip_btn",
        #                  self.helpc("plot_without_asr_dipdip")]),
        #    self.plot_without_asr_dipdip
        #)
        #app(("DOS vs q-mesh", pn.Row(
        #    pn.Column("# DOS vs q-mesh options",
        #              *self.pws("asr", "chneut", "dipdip", "dos_method", "nqsmall_list", "plot_dos_vs_qmesh_btn",
        #                        self.helpc("plot_dos_vs_qmesh")),
        #              ),
        #    self.plot_dos_vs_qmesh)
        #))
        #app(("Quadrupoles", pn.Row(
        #    pn.Column("# Quadrupoles options",
        #              *self.pws("asr", "chneut", "dipdip", "lo_to_splitting", "ndivsm", "dos_method", "plot_phbands_quad_btn",
        #                        self.helpc("plot_phbands_quad")),
        #              ),
        #    self.plot_phbands_quad)
        #))
        #app(("IFCs", pn.Row(
        #    pn.Column("# IFCs options",
        #              *self.pws("asr", "dipdip", "chneut", "plot_ifc_btn",
        #                        self.helpc("on_plot_ifc")),
        #              ),
        #    self.on_plot_ifc)
        #))
        d["Global"] = pn.Row(
            self.pws_col(["## Global options", "units", "mpi_procs", "verbose"]),
            self.get_software_stack()
        )

        if as_dict: return d

        return self.get_template_from_tabs(d, template=kwargs.get("template", None))


class RobotWithFileInput(AbipyParameterized):

    info_str = """
This app allows users to create an AbiPy robot to post-process
a set of ABINIT output files of the same type.
 """

    def __init__(self, **params):

        help_md = pn.pane.Markdown(f"""
## Description

 {self.info_str}
""")

        super().__init__(**params)

        self.main_area = pn.Column(help_md,
                                   self.get_alert_data_transfer(),
                                   sizing_mode="stretch_width")
        self.robot = None
        import os
        top = os.getcwd()
        top = "/Users/gmatteo/git_repos/abipy/abipy/data/refs/mgb2_phonons_nkpt_tsmear"
        top = "~"
        self.file_selector = pnw.FileSelector(top)
        self.robot_files_btn = pnw.Button(name="Load files", button_type='primary', sizing_mode="stretch_width")
        self.robot_files_btn.on_click(self.on_load_files)

    #@depends_on_btn_click("robot_files_btn")
    def on_load_files(self, event):
        if not self.file_selector.value: return
        #self.mpid_err_wdg.object = ""
        #new_abifile = self.get_abifile_from_file_input(self.file_input)
        #print("in on_load_files")
        #print(self.file_selector.value)

        # This should be executed only if server mode: TODO
        #if self.robot is not None:
        #    self.robot.remove()

        from abipy.abilab import abirobot
        self.robot = abirobot(self.file_selector.value)

        self.main_area.objects = [self.robot.get_panel()]

    def get_panel(self):

        # Add help section explaining how to use the filesector. See:
        # https://panel.holoviz.org/reference/widgets/FileSelector.html
        help_md = pn.pane.Markdown("""
Back (◀): Goes to the previous directory

Forward (▶): Returns to the last directory after navigating back

Up (⬆): Goes one directory up.

Address bar: Display the directory to navigate to

Enter (⬇): Navigates to the directory in the address bar

Reload (↻): Reloads the contents of the current directory

To navigate to a subfolder click on a directory in the file selector and then
hit the down arrow (⬇) in the navigation bar.
Files and folders may be selected by selecting them in the browser on the left and moving them
to the right with the arrow buttons.
""")
        col = pn.Column(
            "# Select files (all with the same extension e.g. *_DDB:)",
            self.file_selector,
            pn.Row(self.robot_files_btn, pn.Accordion(("Help", help_md), sizing_mode="stretch_width"),
                   sizing_mode="stretch_width"),
            pn.layout.Divider(),
            sizing_mode="stretch_width")

        main = pn.Column(col, self.main_area, sizing_mode="stretch_width")
        cls, kwds = self.get_abinit_template_cls_kwds()

        return cls(main=main, title="Robot Analyzer", **kwds)



#!/usr/bin/env python

from abipy.electrons.effmass_analyzer import EffMassAnalyzer
import sys

from monty.json import json, MontyEncoder
import numpy as np

#carr = np.array([1j, 2j])
#print(json.dumps(carr, cls=MontyEncoder))

emana = EffMassAnalyzer.from_file(sys.argv[1])
print(emana)
#print(emana.kpoints)

#emana.select_kpoint_band((0, 0, 0), band=7, etol_ev=0.1)
#emana.set_kpoint_band((0, 0, 0), band=3)
#emana.select_band_edges()
#emana.select_cbm()
emana.select_vbm(etol_ev=1e-1)

emana.summarize()
emana.plot_all_segments()
emana.plot_emass(acc=4)

#for segment in emana.segments:
#    segment.get_effmass_line(acc=2)

segment = emana.segments[0]
segment.plot_emass()
print(segment.get_dataframe_with_accuracies())



# flake8: noqa
from __future__ import annotations

import collections

from pymatgen.core.units import Energy
from pymatgen.io.abinit.abiobjects import *


class LujForSpecie(collections.namedtuple("LdauForSpecie", "l u j unit")):
    """
    This object stores the value of l, u, j used for a single atomic specie.
    """
    def __new__(cls, l, u, j, unit):
        """
        Args:
            l: Angular momentum (int or string).
            u: U value
            j: J Value
            unit: Energy unit for u and j.
        """
        l = l
        u, j = Energy(u, unit), Energy(j, unit)
        return super().__new__(cls, l, u, j, unit)


class LdauParams:
    """
    This object stores the parameters for LDA+U calculations with the PAW method
    It facilitates the specification of the U-J parameters in the Abinit input file.
    (see `to_abivars`). The U-J operator will be applied only on the atomic species
    that have been selected by calling `lui_for_symbol`.

    To setup the Abinit variables for a LDA+U calculation in NiO with a
    U value of 5 eV applied on the nickel atoms:

    .. code-block:: python

        luj_params = LdauParams(usepawu=1, structure=nio_structure)
        # Apply U-J on Ni only.
        u = 5.0
        luj_params.luj_for_symbol("Ni", l=2, u=u, j=0.1*u, unit="eV")

        print(luj_params.to_abivars())
    """
    def __init__(self, usepawu, structure):
        """
        Arg:
            usepawu: ABINIT variable `usepawu` defining the LDA+U method.
            structure: |Structure| object.
        """
        self.usepawu = usepawu
        self.structure = structure
        self._params = {}

    @property
    def symbols_by_typat(self) -> list:
        return [specie.symbol for specie in self.structure.species_by_znucl]

    def luj_for_symbol(self, symbol, l, u, j, unit="eV") -> None:
        """
        Args:
            symbol: Chemical symbol of the atoms on which LDA+U should be applied.
            l: Angular momentum.
            u: Value of U.
            j: Value of J.
            unit: Energy unit of U and J.
        """
        if symbol not in self.symbols_by_typat:
            raise ValueError("Symbol %s not in symbols_by_typat:\n%s" % (symbol, self.symbols_by_typat))

        if symbol in self._params:
            raise ValueError("Symbol %s is already present in LdauParams! Cannot overwrite:\n" % symbol)

        self._params[symbol] = LujForSpecie(l=l, u=u, j=j, unit=unit)

    def to_abivars(self) -> dict:
        """Returns a dict with the Abinit variables."""
        lpawu, upawu, jpawu = [], [], []

        for symbol in self.symbols_by_typat:
            p = self._params.get(symbol, None)

            if p is not None:
                l, u, j = p.l, p.u.to("eV"), p.j.to("eV")
            else:
                l, u, j = -1, 0, 0

            lpawu.append(int(l))
            upawu.append(float(u))
            jpawu.append(float(j))

        # convert upawu and jpaw to string so that we can use
        # eV unit in the Abinit input file (much more readable).
        return dict(
            usepawu=self.usepawu,
            lpawu=" ".join(map(str, lpawu)),
            upawu=" ".join(map(str, upawu)) + " eV",
            jpawu=" ".join(map(str, jpawu)) + " eV")


class LexxParams:
    """
    This object stores the parameters for local exact exchange calculations with the PAW method
    It facilitates the specification of the LEXX parameters in the Abinit input file.
    (see `to_abivars`). The LEXX operator will be applied only on the atomic species
    that have been selected by calling `lexx_for_symbol`.

    To perform a LEXX calculation for NiO in which the LEXX is computed only for the l=2
    channel of the nickel atoms:

    .. code-block:: python

        lexx_params = LexxParams(nio_structure)
        lexx_params.lexx_for_symbol("Ni", l=2)

        print(lexc_params.to_abivars())
    """
    def __init__(self, structure):
        """
        Arg:
            structure: |Structure| object.
        """
        self.structure = structure
        self._lexx_for_symbol = {}

    @property
    def symbols_by_typat(self) -> list:
        return [specie.symbol for specie in self.structure.species_by_znucl]

    def lexx_for_symbol(self, symbol, l) -> None:
        """
        Enable LEXX for the given chemical symbol and the angular momentum l

        Args:
            symbol: Chemical symbol of the atoms on which LEXX should be applied.
            l: Angular momentum.
        """
        if symbol not in self.symbols_by_typat:
            err_msg = "Symbol %s not in symbols_by_typat:\n%s" % (symbol, self.symbols_by_typat)
            raise ValueError(err_msg)

        if symbol in self._lexx_for_symbol:
            raise ValueError("Symbol %s is already present in LdauParams! Cannot overwrite:" % symbol)

        self._lexx_for_symbol[symbol] = l

    def to_abivars(self) -> dict:
        """Returns a dict with the Abinit variables."""
        lexx_typat = []

        for symbol in self.symbols_by_typat:
            l = self._lexx_for_symbol.get(symbol, -1)
            lexx_typat.append(int(l))

        return dict(useexexch=1, lexexch=" ".join(map(str, lexx_typat)))



#!/usr/bin/env python
r"""
LM-fatbands
===========

This example shows how to plot the LM-projected fatbands.
The FATBANDS file must have benn produced with prtdos 3 and prtdosm 1.
"""
import abipy.abilab as abilab
import abipy.data as abidata

fbnc_kpath = abilab.abiopen(abidata.ref_file("ni_kpath_FATBANDS.nc"))
print(fbnc_kpath)

# NC files have contributions up to L = 4 (g channel)
# but here we are intererested in s,p,d terms only so we use the optional argument lmax
lmax = 2

# we are not interested in a small energy window around the Fermi level.
elims = [-1.5, +1]

# and a subset of bands (remember that in python we start to count from 0)
blist = list(range(4, 10))

# Plot fatbands with LM character up to lmax.
# The grid contains (lmax + 1) columns, each column has (2l + 1) subplots
# corresponding to the LM character for M in [-l, -l-1, ... 0, 1, +l].
fbnc_kpath.plot_fatbands_mview(iatom=0, fact=1.5, lmax=lmax,
                               ylims=elims, blist=list(range(4, 10)),
                               title="LM fatbands for atom index 0")

fbnc_kpath.close()



#!/usr/bin/env python
r"""
Convergence studiy with different k-meshes
==========================================

In this example,
"""
import sys
import os

import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata


def build_flow(options):
    """
    Build and return a flow for GS properties with different k-meshes
    """
    # Set working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    structure = abilab.Structure.from_file(abidata.cif_file("gan2.cif"))
    pseudos = abidata.pseudos("Ga.oncvpsp", "N.oncvpsp")

    scf_input = abilab.AbinitInput(structure, pseudos)

    scf_input.set_vars(
        ecut=15,       # Too low.
        nstep=50,      # Increase default
        tolvrs=1e-8,
    )

    flow = flowtk.Flow(workdir=options.workdir)

    from abipy.flowtk.gs_works import GsKmeshConvWork
    nksmall_list = [2, 4, 6, 8]
    flow.register_work(GsKmeshConvWork.from_scf_input(scf_input, nksmall_list))

    return flow



# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())




#!/usr/bin/env python
r"""
Effective masses with finite diff
=================================

This example shows how to use finite differences to compute
effective masses and plot the data.
"""

#%%
# Initialize the EffMassAnalyzer from a GSR file with energies
# computed along segments passing through the k-points of interest.
#
# Clearly the accuracy of the finite difference results strongly depends
# on the spacing between the points along the path.
# The GSR file used in this example is not accurate enough!

import abipy.data as abidata
from abipy.electrons.effmass_analyzer import EffMassAnalyzer

emana = EffMassAnalyzer.from_file(abidata.ref_file("si_nscf_GSR.nc"))
print(emana)

#%%
# Before computing the effective masses, we have to select the k-points.

emana.select_vbm()

# Alternatively, one can use:

#emana.select_cbm()
#emana.select_band_edges()

# or the most flexible API:

#emana.select_kpoint_band((0, 0, 0), band=3, spin=0, degtol_ev=0.1)

#%%
# Compute effective masses with different accuracies and print results in tabular format.
# Useful to understand if results are sensitive to the number of points in the finite difference.
# Note however that numerical differentiation is performed with the same delta_k step.

emana.summarize()

#%%
# Print the results to terminal with:

# extract segment.
segment = emana.segments[0]
#repr(segment); str(segment)
#assert segment.to_string(verbose=2)
df = segment.get_dataframe_with_accuracies(acc_list=(2, 4))
print(df)

#assert len(emana.segments) == 1
#for segment in emana.segments[0]:
#    segment.get_effmass_line(acc=2)

#%%
# Plot electronic dispersion and quadratic approximant based on the
# effective masses computed along each segment.

emana.plot_emass(acc=4)

#emana.plot_all_segments()
#emana.segments[0].plot_emass()



#!/usr/bin/env python
r"""
Screening Flow with q-point parallelism
=======================================

This example shows how to compute the SCR file by splitting the calculation of the SCR file
over q-points with the input variables nqptdm and qptdm.
"""
from __future__ import division, print_function, unicode_literals, absolute_import

import sys
import os
import abipy.abilab as abilab
import abipy.data as data

from abipy import flowtk


def all_inputs(paral_kgb=1):
    """
    Build the input files of the calculation.
    Returns: gs_input, nscf_input, scr_input, sigma_input
    """
    ecut = ecutwfn = 4
    global_vars = dict(
        ecut=ecut,
        istwfk="*1",
        paral_kgb=paral_kgb,
    )

    multi = abilab.MultiDataset(structure=data.cif_file("si.cif"),
                                pseudos=data.pseudos("14si.pspnc"), ndtset=4)
    multi.set_vars(global_vars)

    gs, nscf, scr, sigma = multi.split_datasets()

    # This grid is the most economical, but does not contain the Gamma point.
    gs_kmesh = dict(
        ngkpt=[2, 2, 2],
        shiftk=[0.5, 0.5, 0.5,
                0.5, 0.0, 0.0,
                0.0, 0.5, 0.0,
                0.0, 0.0, 0.5]
    )

    # This grid contains the Gamma point, which is the point at which
    # we will compute the (direct) band gap.
    gw_kmesh = dict(
        ngkpt=[2, 2, 2],
        shiftk=[0.0, 0.0, 0.0,
                0.0, 0.5, 0.5,
                0.5, 0.0, 0.5,
                0.5, 0.5, 0.0]
    )

    # Dataset 1 (GS run)
    gs.set_kmesh(**gs_kmesh)
    gs.set_vars(tolvrs=1e-6, nband=4)

    # Dataset 2 (NSCF run)
    # Here we select the second dataset directly with the syntax inp[2]
    nscf.set_kmesh(**gw_kmesh)
    nscf.set_vars(iscf=-2,
                  tolwfr=1e-10,
                  nband=15,
                  nbdbuf=5,
    )

    # Dataset3: Calculation of the screening.
    scr.set_kmesh(**gw_kmesh)
    scr.set_vars(
        optdriver=3,
        nband=6,
        ecutwfn=ecutwfn,
        symchi=1,
        inclvkb=0,
        ecuteps=2.0,
    )

    # Dataset4: Calculation of the Self-Energy matrix elements (GW corrections)
    sigma.set_kmesh(**gw_kmesh)
    sigma.set_vars(
            optdriver=4,
            nband=8,
            ecutwfn=ecutwfn,
            ecuteps=2.0,
            ecutsigx=2.0,
            symsigma=1,
    )

    kptgw = [
         -2.50000000E-01, -2.50000000E-01,  0.00000000E+00,
         -2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
          5.00000000E-01,  5.00000000E-01,  0.00000000E+00,
         -2.50000000E-01,  5.00000000E-01,  2.50000000E-01,
          5.00000000E-01,  0.00000000E+00,  0.00000000E+00,
          0.00000000E+00,  0.00000000E+00,  0.00000000E+00,
    ]

    bdgw = [1, 8]
    sigma.set_kptgw(kptgw, bdgw)

    return multi.split_datasets()


def build_flow(options):
    """
    Construct the flow for G0W0 calculations.
    The calculation of the SCR file is split into nqptdm independent calculations.
    Partial SCR files are then merged with mrgddb before starting the sigma run.
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        __file__ = os.path.join(os.getcwd(), "run_qptdmscr.py")
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    # Build the input files for GS, NSCF, SCR and SIGMA runs.
    gs, nscf, scr_input, sigma_input = all_inputs()

    # Construct the flow.
    return flowtk.G0W0WithQptdmFlow(options.workdir, gs, nscf, scr_input, sigma_input,
                                    manager=options.manager)


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""
Tools for writing cube files.
See http://paulbourke.net/dataformats/cube/ and http://www.gaussian.com/g_tech/g_ur/u_cubegen.htm
"""
from __future__ import annotations

import numpy as np

from pymatgen.core.lattice import Lattice
from pymatgen.core.sites import PeriodicSite
from pymatgen.core.units import bohr_to_angstrom


__all__ = [
    "cube_write_structure_mesh",
    "cube_write_data",
]


def cube_write_structure_mesh(file, structure, mesh) -> None:
    fwrite = file.write
    fwrite("Density generated from abipy\n")
    fwrite("in the cube file format\n")
    dvx = mesh.dvx / bohr_to_angstrom
    dvy = mesh.dvy / bohr_to_angstrom
    dvz = mesh.dvz / bohr_to_angstrom
    fwrite('{:4d} {:.6f} {:.6f} {:.6f}\n'.format(len(structure), 0.0, 0.0, 0.0))
    fwrite('{:4d} {:.6f} {:.6f} {:.6f}\n'.format(mesh.nx, dvx[0], dvx[1], dvx[2]))
    fwrite('{:4d} {:.6f} {:.6f} {:.6f}\n'.format(mesh.ny, dvy[0], dvy[1], dvy[2]))
    fwrite('{:4d} {:.6f} {:.6f} {:.6f}\n'.format(mesh.nz, dvz[0], dvz[1], dvz[2]))
    for site in structure:
        cc = site.coords / bohr_to_angstrom
        fwrite('{:d} {:.10f} {:.10f} {:.10f} {:.10f}\n'.format(site.specie.Z, 0.0, cc[0], cc[1], cc[2]))


def cube_write_data(file, data, mesh) -> None:
    fwrite = file.write
    data_bohrs = data * (bohr_to_angstrom ** 3)
    for ix in range(mesh.nx):
        for iy in range(mesh.ny):
            for iz in range(mesh.nz):
                fwrite('{:.5e}\n'.format(data_bohrs[ix, iy, iz]))


def cube_read_structure_mesh_data(filepath: str) -> tuple:
    with open(filepath, 'r') as fh:
        # The two first lines are comments
        for ii in range(2):
            fh.readline()
        # Number of atoms
        natoms = int(fh.readline().split()[0])
        # The next three lines give the mesh and the vectors
        sp = fh.readline().split()
        nx = int(sp[0])
        dvx = np.array([float(sp[ii]) for ii in range(1, 4)]) * bohr_to_angstrom
        sp = fh.readline().split()
        ny = int(sp[0])
        dvy = np.array([float(sp[ii]) for ii in range(1, 4)]) * bohr_to_angstrom
        sp = fh.readline().split()
        nz = int(sp[0])
        dvz = np.array([float(sp[ii]) for ii in range(1, 4)]) * bohr_to_angstrom
        uc_matrix = np.array([nx*dvx, ny*dvy, nz*dvz])
        sites = []
        lattice = Lattice(uc_matrix)
        for ii in range(natoms):
            sp = fh.readline().split()
            cc = np.array([float(sp[ii]) for ii in range(2, 5)]) * bohr_to_angstrom
            sites.append(PeriodicSite(int(sp[0]), coords=cc, lattice=lattice, to_unit_cell=False,
                                      coords_are_cartesian=True))
        data = np.zeros((nx, ny, nz))
        ii = 0
        for line in fh:
            for val in line.split():
                data[ii // (ny * nz), (ii // nz) % ny, ii % nz] = float(val)
                ii += 1
        data = data / (bohr_to_angstrom ** 3)
        if ii != nx*ny*nz:
            raise ValueError('Wrong number of data points ...')
        from abipy.core.structure import Structure
        structure = Structure.from_sites(sites=sites)
        from abipy.core.mesh3d import Mesh3D
        mesh = Mesh3D(shape=[nx, ny, nz], vectors=uc_matrix)

        return structure, mesh, data
